from pathlib import Path
from cg.io.json import write_json
from cg.meta.workflow.mutant.quality_controller.models import (
    CaseQualityResult,
    MutantReport,
    SamplesQualityResults,
)
from cg.meta.workflow.mutant.quality_controller.result_logger_utils import (
    get_samples_results_message,
)


def write_report(
    case_qc_report_path: Path,
    case_quality_result: CaseQualityResult,
    samples_quality_results: SamplesQualityResults,
) -> None:
    summary: str = get_summary(
        case_quality_result=case_quality_result,
        samples_quality_results=samples_quality_results,
    )
    report = MutantReport(
        summary=summary,
        case=case_quality_result.model_dump(),
        samples=samples_quality_results.model_dump(),
    )

    write_json(file_path=case_qc_report_path, content=report.model_dump())


def get_summary(
    case_quality_result: CaseQualityResult,
    samples_quality_results: SamplesQualityResults,
) -> str:
    case_summary: str = "Case passed QC. " if case_quality_result.passes_qc else "Case failed QC. "
    sample_summary: str = get_samples_results_message(
        samples_quality_results=samples_quality_results
    )
    summary = case_summary + sample_summary
    return summary



import pytest

from cg.services.deliver_files.file_fetcher.models import SampleFile
from cg.services.deliver_files.file_formatter.destination.models import FormattedFile
from cg.services.deliver_files.file_formatter.path_name.flat_structure import (
    FlatStructurePathFormatter,
)
from cg.services.deliver_files.file_formatter.path_name.nested_structure import (
    NestedStructurePathFormatter,
)


@pytest.mark.parametrize(
    "sample_files,expected_formatted_files,path_name_formatter",
    [
        (
            "expected_moved_analysis_sample_delivery_files",
            "expected_formatted_analysis_sample_files",
            NestedStructurePathFormatter(),
        ),
        (
            "expected_moved_analysis_sample_delivery_files",
            "expected_flat_formatted_analysis_sample_files",
            FlatStructurePathFormatter(),
        ),
    ],
)
def test_path_name_formatters(
    sample_files: list[SampleFile],
    expected_formatted_files: list[FormattedFile],
    path_name_formatter,
    request,
):
    # GIVEN existing sample files and a sample file formatter
    sample_files: list[SampleFile] = request.getfixturevalue(sample_files)
    expected_formatted_files: list[FormattedFile] = request.getfixturevalue(
        expected_formatted_files
    )

    # WHEN formatting the sample files
    formatted_files: list[FormattedFile] = [
        FormattedFile(
            formatted_path=path_name_formatter.format_file_path(
                file_path=sample_file.file_path,
                provided_name=sample_file.sample_name,
                provided_id=sample_file.sample_id,
            ),
            original_path=sample_file.file_path,
        )
        for sample_file in sample_files
    ]

    # THEN the sample files should be formatted
    assert formatted_files == expected_formatted_files



"""Tests the CLI for generating delivery reports."""

import pytest
from _pytest.fixtures import FixtureRequest
from click.testing import CliRunner, Result

from cg.cli.generate.delivery_report.base import generate_delivery_report
from cg.constants import EXIT_SUCCESS, Workflow
from cg.models.cg_config import CGConfig


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_generate_delivery_report(
    request: FixtureRequest, workflow: Workflow, cli_runner: CliRunner
) -> None:
    """Test command to generate delivery report."""

    # GIVEN a delivery report context
    delivery_report_context: CGConfig = request.getfixturevalue(
        f"{workflow}_delivery_report_context"
    )

    # GIVEN a case ID
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # WHEN calling delivery report generation
    result: Result = cli_runner.invoke(
        generate_delivery_report, [case_id], obj=delivery_report_context
    )

    # THEN the generation command should have successfully executed
    assert result.exit_code == EXIT_SUCCESS


@pytest.mark.parametrize("workflow", [Workflow.RAREDISEASE, Workflow.RNAFUSION])
def test_generate_delivery_report_dry_run(
    request: FixtureRequest, workflow: Workflow, cli_runner: CliRunner
) -> None:
    """Test dry run command to generate delivery report."""

    # GIVEN a delivery report context
    delivery_report_context: CGConfig = request.getfixturevalue(
        f"{workflow}_delivery_report_context"
    )

    # GIVEN a case ID
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # WHEN calling delivery report generation with a dry option to print the HTML content to the terminal
    result: Result = cli_runner.invoke(
        generate_delivery_report, [case_id, "--dry-run"], obj=delivery_report_context
    )

    # THEN the generation command should have successfully executed
    assert result.exit_code == EXIT_SUCCESS
    assert "<!DOCTYPE html>" in result.output



"""Test coverage analysis API for human clinical sequencing data."""

from unittest.mock import Mock

import pytest
import requests
from pytest_mock import MockFixture

from cg.clients.chanjo2.client import Chanjo2APIClient
from cg.clients.chanjo2.models import (
    CoverageMetrics,
    CoveragePostRequest,
    CoveragePostResponse,
)
from cg.exc import Chanjo2RequestError, Chanjo2ResponseError


def test_post_coverage_success(
    chanjo2_api_client: Chanjo2APIClient,
    coverage_post_request: CoveragePostRequest,
    coverage_post_response: CoveragePostResponse,
    coverage_post_response_success: Mock,
    mocker: MockFixture,
):
    """Test successful POST request coverage extraction."""

    # GIVEN a mocked POST request
    mocker.patch.object(requests, "post", return_value=coverage_post_response_success)

    # WHEN getting the coverage data
    sample_coverage: CoveragePostResponse = chanjo2_api_client.get_coverage(coverage_post_request)

    # THEN the returned coverage data should match the expected one
    assert sample_coverage == coverage_post_response


def test_get_coverage_http_error(
    chanjo2_api_client: Chanjo2APIClient,
    coverage_post_request: CoveragePostRequest,
    coverage_post_response_http_error: Mock,
    mocker: MockFixture,
):
    """Test the handling of an HTTP error when getting coverage."""

    # GIVEN a mocked POST request raising an exception
    mocker.patch.object(requests, "post", return_value=coverage_post_response_http_error)

    # WHEN getting the coverage data

    # THEN a Chanjo2 request error should have been raised
    with pytest.raises(Chanjo2RequestError):
        chanjo2_api_client.get_coverage(coverage_post_request)


def test_get_coverage_invalid_values(
    chanjo2_api_client: Chanjo2APIClient,
    coverage_post_request: CoveragePostRequest,
    coverage_post_response_invalid_values: Mock,
    mocker: MockFixture,
):
    """Test handling of a validation error when getting coverage."""

    # GIVEN a mocked POST request returning a JSON with invalid values
    mocker.patch.object(requests, "post", return_value=coverage_post_response_invalid_values)

    # WHEN getting the coverage data

    # THEN a Chanjo2 response error should have been raised
    with pytest.raises(Chanjo2ResponseError):
        chanjo2_api_client.get_coverage(coverage_post_request)


def test_get_coverage_invalid_attributes(
    chanjo2_api_client: Chanjo2APIClient,
    coverage_post_request: CoveragePostRequest,
    coverage_post_response_invalid_attributes: Mock,
    mocker: MockFixture,
):
    """Test coverage POST request returning an invalid response with incorrect attributes."""

    # GIVEN a mocked POST request returning a JSON with incorrect attributes
    mocker.patch.object(requests, "post", return_value=coverage_post_response_invalid_attributes)

    # WHEN getting the coverage data

    # THEN a Chanjo2 response error should have been raised
    with pytest.raises(Chanjo2ResponseError):
        chanjo2_api_client.get_coverage(coverage_post_request)


def test_get_coverage_empty_response(
    chanjo2_api_client: Chanjo2APIClient,
    coverage_post_request: CoveragePostRequest,
    coverage_post_response_empty: Mock,
    mocker: MockFixture,
):
    """Test coverage POST request returning an empty response."""

    # GIVEN a mocked POST request returning an empty JSON
    mocker.patch.object(requests, "post", return_value=coverage_post_response_empty)

    # WHEN getting the coverage data

    # THEN a Chanjo2 response error should have been raised
    with pytest.raises(Chanjo2ResponseError):
        chanjo2_api_client.get_coverage(coverage_post_request)


def test_get_sample_coverage_metrics(
    sample_id: str,
    coverage_post_response: CoveragePostResponse,
    coverage_post_response_json: dict[str, dict],
):
    """Test sample coverage extraction from a coverage POST response."""

    # GIVEN a mocked POST response

    # WHEN getting the coverage data for a specific sample ID
    coverage_metrics: CoverageMetrics = coverage_post_response.get_sample_coverage_metrics(
        sample_id
    )

    # THEN the returned coverage metrics should match the expected one
    assert coverage_metrics.model_dump() == coverage_post_response_json[sample_id]


def test_get_sample_coverage_metrics_invalid_sample_id(
    invalid_sample_id: str, coverage_post_response: CoveragePostResponse
):
    """Test sample coverage extraction from a coverage POST response providing an invalid sample."""

    # GIVEN a mocked POST response

    # WHEN getting the coverage data for a specific sample ID

    # THEN a validation error should be raised
    with pytest.raises(ValueError):
        coverage_post_response.get_sample_coverage_metrics(invalid_sample_id)



"""Unified interface to handle sample submissions.

This interface will update information in Status and/or LIMS as required.

The normal entry for information is through the REST API which will pass a JSON
document with all information about samples in the submission. The input will
be validated and if passing all checks be accepted as new samples.
"""

import logging

from cg.apps.lims import LimsAPI
from cg.meta.orders.ticket_handler import TicketHandler
from cg.models.orders.order import OrderIn, OrderType
from cg.services.orders.submitters.order_submitter_registry import OrderSubmitterRegistry
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class OrdersAPI:
    """Orders API for accepting new samples into the system."""

    def __init__(
        self,
        lims: LimsAPI,
        status: Store,
        ticket_handler: TicketHandler,
        submitter_registry: OrderSubmitterRegistry,
    ):
        super().__init__()
        self.lims = lims
        self.status = status
        self.ticket_handler = ticket_handler
        self.submitter_registry = submitter_registry

    def submit(self, project: OrderType, order_in: OrderIn, user_name: str, user_mail: str) -> dict:
        """Submit a batch of samples.

        Main entry point for the class towards interfaces that implements it.
        """
        submit_handler = self.submitter_registry.get_order_submitter(project)
        submit_handler.order_validation_service.validate_order(order_in)
        ticket_number = self.ticket_handler.create_ticket(
            order=order_in, user_name=user_name, user_mail=user_mail, project=project
        )
        order_in.ticket = ticket_number
        return submit_handler.submit_order(order_in=order_in)



from pathlib import Path

import pytest

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn


@pytest.fixture(scope="session")
def balsamic_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example cancer order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "balsamic.json")
    )


@pytest.fixture(scope="session")
def fastq_order_to_submit(cgweb_orders_dir) -> dict:
    """Load an example FASTQ order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "fastq.json")
    )


@pytest.fixture(scope="session")
def metagenome_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example metagenome order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "metagenome.json")
    )


@pytest.fixture(scope="session")
def microbial_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example microbial order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "microsalt.json")
    )


@pytest.fixture
def microbial_fastq_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example microbial order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "microbial_fastq.json")
    )


@pytest.fixture(scope="session")
def mip_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example MIP order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "mip.json")
    )


@pytest.fixture(scope="session")
def mip_rna_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example RNA order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "mip_rna.json")
    )


@pytest.fixture(scope="session")
def pacbio_order_to_submit(cgweb_orders_dir) -> dict:
    """Load an example PacBio order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "pacbio.json")
    )


@pytest.fixture(scope="session")
def rnafusion_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example RNA order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "rnafusion.json")
    )


@pytest.fixture(scope="session")
def rml_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example RML order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "rml.json")
    )


@pytest.fixture(scope="session")
def sarscov2_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example sarscov2 order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "sarscov2.json")
    )


@pytest.fixture(scope="session")
def tomte_order_to_submit(cgweb_orders_dir: Path) -> dict:
    """Load an example TOMTE order."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(cgweb_orders_dir, "tomte.json")
    )


@pytest.fixture(scope="session")
def all_orders_to_submit(
    balsamic_order_to_submit: dict,
    fastq_order_to_submit: dict,
    metagenome_order_to_submit: dict,
    microbial_order_to_submit: dict,
    mip_order_to_submit: dict,
    mip_rna_order_to_submit: dict,
    pacbio_order_to_submit: dict,
    rml_order_to_submit: dict,
    rnafusion_order_to_submit: dict,
    sarscov2_order_to_submit: dict,
) -> dict[str, OrderIn]:
    """Returns a dict of parsed order for each order type."""
    return {
        OrderType.BALSAMIC: OrderIn.parse_obj(balsamic_order_to_submit, project=OrderType.BALSAMIC),
        OrderType.FASTQ: OrderIn.parse_obj(fastq_order_to_submit, project=OrderType.FASTQ),
        OrderType.FLUFFY: OrderIn.parse_obj(rml_order_to_submit, project=OrderType.FLUFFY),
        OrderType.METAGENOME: OrderIn.parse_obj(
            metagenome_order_to_submit, project=OrderType.METAGENOME
        ),
        OrderType.MICROSALT: OrderIn.parse_obj(
            microbial_order_to_submit, project=OrderType.MICROSALT
        ),
        OrderType.MIP_DNA: OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA),
        OrderType.MIP_RNA: OrderIn.parse_obj(mip_rna_order_to_submit, project=OrderType.MIP_RNA),
        OrderType.PACBIO_LONG_READ: OrderIn.parse_obj(
            pacbio_order_to_submit, project=OrderType.PACBIO_LONG_READ
        ),
        OrderType.RML: OrderIn.parse_obj(rml_order_to_submit, project=OrderType.RML),
        OrderType.RNAFUSION: OrderIn.parse_obj(
            rnafusion_order_to_submit, project=OrderType.RNAFUSION
        ),
        OrderType.SARS_COV_2: OrderIn.parse_obj(
            sarscov2_order_to_submit, project=OrderType.SARS_COV_2
        ),
    }



import copy
import logging
import re
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator

from dateutil.parser import parse
from pydantic.v1 import ValidationError

from cg.constants import Workflow
from cg.constants.constants import (
    CaseActions,
    FileExtensions,
    FileFormat,
    GenomeVersion,
    MultiQC,
    WorkflowManager,
)
from cg.constants.gene_panel import GenePanelGenomeBuild
from cg.constants.nextflow import NFX_WORK_DIR
from cg.constants.nf_analysis import NfTowerStatus
from cg.constants.tb import AnalysisStatus
from cg.exc import CgError, HousekeeperStoreError, MetricsQCError
from cg.io.controller import ReadFile, WriteFile
from cg.io.json import read_json
from cg.io.txt import concat_txt, write_txt
from cg.io.yaml import read_yaml, write_yaml_nextflow_style
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.nf_handlers import NextflowHandler, NfTowerHandler
from cg.meta.workflow.utils.genome_build_helpers import get_genome_build
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import (
    MetricsBase,
    MetricsDeliverablesCondition,
    MultiqcDataJson,
)
from cg.models.fastq import FastqFileMeta
from cg.models.nf_analysis import (
    FileDeliverable,
    NfCommandArgs,
    WorkflowDeliverables,
    WorkflowParameters,
)
from cg.store.models import Analysis, Case, CaseSample, Sample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class NfAnalysisAPI(AnalysisAPI):
    """Parent class for handling NF-core analyses."""

    def __init__(self, config: CGConfig, workflow: Workflow):
        super().__init__(workflow=workflow, config=config)
        self.workflow: Workflow = workflow
        self.root_dir: str | None = None
        self.workflow_bin_path: str | None = None
        self.references: str | None = None
        self.profile: str | None = None
        self.conda_env: str | None = None
        self.conda_binary: str | None = None
        self.platform: str | None = None
        self.params: str | None = None
        self.config: str | None = None
        self.resources: str | None = None
        self.tower_binary_path: str | None = None
        self.tower_workflow: str | None = None
        self.account: str | None = None
        self.email: str | None = None
        self.compute_env_base: str | None = None
        self.revision: str | None = None
        self.nextflow_binary_path: str | None = None

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def process(self):
        if not self._process:
            self._process = Process(
                binary=self.tower_binary_path,
            )
        return self._process

    @process.setter
    def process(self, process: Process):
        self._process = process

    @property
    def use_read_count_threshold(self) -> bool:
        """Defines whether the threshold for adequate read count should be passed for all samples
        when determining if the analysis for a case should be automatically started."""
        return True

    @property
    def sample_sheet_headers(self) -> list[str]:
        """Headers for sample sheet."""
        raise NotImplementedError

    @property
    def is_multiqc_pattern_search_exact(self) -> bool:
        """Return True if only exact pattern search is allowed to collect metrics information from MultiQC file.
        If false, pattern must be present but does not need to be exact."""
        return False

    @property
    def is_gene_panel_required(self) -> bool:
        """Return True if a gene panel needs to be created using the information in StatusDB and exporting it from Scout."""
        return False

    @property
    def is_managed_variants_required(self) -> bool:
        """Return True if a managed variant export needs to be exported it from Scout."""
        return False

    def get_profile(self, profile: str | None = None) -> str:
        """Get NF profiles."""
        return profile or self.profile

    def get_workflow_manager(self) -> str:
        """Get workflow manager from Tower."""
        return WorkflowManager.Tower.value

    def get_workflow_version(self, case_id: str) -> str:
        """Get workflow version from config."""
        return self.revision

    def get_built_workflow_parameters(self, case_id: str) -> WorkflowParameters:
        """Return workflow parameters."""
        raise NotImplementedError

    def get_nextflow_config_content(self, case_id: str) -> str:
        """Return nextflow config content."""
        config_files_list: list[str] = [
            self.platform,
            self.config,
            self.resources,
        ]
        extra_parameters_str: list[str] = [
            self.set_cluster_options(case_id=case_id),
        ]
        return concat_txt(
            file_paths=config_files_list,
            str_content=extra_parameters_str,
        )

    def get_case_path(self, case_id: str) -> Path:
        """Path to case working directory."""
        return Path(self.root_dir, case_id)

    def get_sample_sheet_path(self, case_id: str) -> Path:
        """Path to sample sheet."""
        return Path(self.get_case_path(case_id), f"{case_id}_samplesheet").with_suffix(
            FileExtensions.CSV
        )

    def get_compute_env(self, case_id: str) -> str:
        """Get the compute environment for the head job based on the case priority."""
        return f"{self.compute_env_base}-{self.get_slurm_qos_for_case(case_id=case_id)}"

    def get_nextflow_config_path(
        self, case_id: str, nextflow_config: Path | str | None = None
    ) -> Path:
        """Path to nextflow config file."""
        if nextflow_config:
            return Path(nextflow_config).absolute()
        return Path((self.get_case_path(case_id)), f"{case_id}_nextflow_config").with_suffix(
            FileExtensions.JSON
        )

    def get_job_ids_path(self, case_id: str) -> Path:
        """Return the path to a Trailblazer config file containing Tower IDs."""
        return Path(self.root_dir, case_id, "tower_ids").with_suffix(FileExtensions.YAML)

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Path to deliverables file for a case."""
        return Path(self.get_case_path(case_id), f"{case_id}_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def get_metrics_deliverables_path(self, case_id: str) -> Path:
        """Return a path where the <case>_metrics_deliverables.yaml file should be located."""
        return Path(self.root_dir, case_id, f"{case_id}_metrics_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def get_params_file_path(self, case_id: str, params_file: Path | None = None) -> Path:
        """Return parameters file or a path where the default parameters file for a case id should be located."""
        if params_file:
            return Path(params_file).absolute()
        return Path((self.get_case_path(case_id)), f"{case_id}_params_file").with_suffix(
            FileExtensions.YAML
        )

    def create_case_directory(self, case_id: str, dry_run: bool = False) -> None:
        """Create case directory."""
        if not dry_run:
            Path(self.get_case_path(case_id=case_id)).mkdir(parents=True, exist_ok=True)

    def get_log_path(self, case_id: str, workflow: str) -> Path:
        """Path to NF log."""
        launch_time: str = datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
        return Path(
            self.get_case_path(case_id),
            f"{case_id}_{workflow}_nextflow_{launch_time}",
        ).with_suffix(FileExtensions.LOG)

    def get_workdir_path(self, case_id: str, work_dir: Path | None = None) -> Path:
        """Path to NF work directory."""
        if work_dir:
            return work_dir.absolute()
        return Path(self.get_case_path(case_id), NFX_WORK_DIR)

    def get_gene_panels_path(self, case_id: str) -> Path:
        """Path to gene panels bed file exported from Scout."""
        return Path(self.get_case_path(case_id=case_id), "gene_panels").with_suffix(
            FileExtensions.BED
        )

    def set_cluster_options(self, case_id: str) -> str:
        return f'process.clusterOptions = "-A {self.account} --qos={self.get_slurm_qos_for_case(case_id=case_id)}"\n'

    @staticmethod
    def extract_read_files(
        metadata: list[FastqFileMeta], forward_read: bool = False, reverse_read: bool = False
    ) -> list[str]:
        """Extract a list of fastq file paths for either forward or reverse reads."""
        if forward_read and not reverse_read:
            read_direction = 1
        elif reverse_read and not forward_read:
            read_direction = 2
        else:
            raise ValueError("Either forward or reverse needs to be specified")
        sorted_metadata: list = sorted(metadata, key=lambda k: k.path)
        return [
            fastq_file.path
            for fastq_file in sorted_metadata
            if fastq_file.read_direction == read_direction
        ]

    def get_paired_read_paths(self, sample=Sample) -> tuple[list[str], list[str]]:
        """Returns a tuple of paired fastq file paths for the forward and reverse read."""
        sample_metadata: list[FastqFileMeta] = self.gather_file_metadata_for_sample(sample=sample)
        fastq_forward_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, forward_read=True
        )
        fastq_reverse_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, reverse_read=True
        )
        return fastq_forward_read_paths, fastq_reverse_read_paths

    def get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        raise NotImplementedError

    def get_sample_sheet_content(self, case_id: str) -> list[list[Any]]:
        """Return formatted information required to build a sample sheet for a case.
        This contains information for all samples linked to the case."""
        sample_sheet_content: list = []
        case: Case = self.get_validated_case(case_id)
        LOG.info(f"Samples linked to case {case_id}: {len(case.links)}")
        LOG.debug("Getting sample sheet information")
        for link in case.links:
            sample_sheet_content.extend(self.get_sample_sheet_content_per_sample(case_sample=link))
        return sample_sheet_content

    def verify_sample_sheet_exists(self, case_id: str, dry_run: bool = False) -> None:
        """Raise an error if sample sheet file is not found."""
        if not dry_run and not Path(self.get_sample_sheet_path(case_id=case_id)).exists():
            raise ValueError(f"No config file found for case {case_id}")

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        """Raise an error if a deliverable file is not found."""
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    def write_params_file(self, case_id: str, replaced_workflow_parameters: dict = None) -> None:
        """Write params-file for analysis."""
        LOG.debug("Writing parameters file")
        if replaced_workflow_parameters:
            write_yaml_nextflow_style(
                content=replaced_workflow_parameters,
                file_path=self.get_params_file_path(case_id=case_id),
            )
        else:
            self.get_params_file_path(case_id=case_id).touch()

    @staticmethod
    def write_sample_sheet(
        content: list[list[Any]],
        file_path: Path,
        header: list[str],
    ) -> None:
        """Write sample sheet CSV file."""
        LOG.debug("Writing sample sheet")
        if header:
            content.insert(0, header)
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.CSV,
            file_path=file_path,
        )

    @staticmethod
    def write_deliverables_file(
        deliverables_content: dict, file_path: Path, file_format=FileFormat.YAML
    ) -> None:
        """Write deliverables file."""
        WriteFile.write_file_from_content(
            content=deliverables_content, file_format=file_format, file_path=file_path
        )

    def write_trailblazer_config(self, case_id: str, tower_id: str) -> None:
        """Write Tower IDs to a file used as the Trailblazer config."""
        config_path: Path = self.get_job_ids_path(case_id=case_id)
        LOG.info(f"Writing Tower ID to {config_path.as_posix()}")
        WriteFile.write_file_from_content(
            content={case_id: [tower_id]},
            file_format=FileFormat.YAML,
            file_path=config_path,
        )

    def create_sample_sheet(self, case_id: str, dry_run: bool) -> None:
        """Create sample sheet for a case."""
        sample_sheet_content: list[list[Any]] = self.get_sample_sheet_content(case_id=case_id)
        if not dry_run:
            self.write_sample_sheet(
                content=sample_sheet_content,
                file_path=self.get_sample_sheet_path(case_id=case_id),
                header=self.sample_sheet_headers,
            )

    def create_params_file(self, case_id: str, dry_run: bool) -> None:
        """Create parameters file for a case."""
        LOG.debug("Getting parameters information built on-the-fly")
        built_workflow_parameters: dict | None = self.get_built_workflow_parameters(
            case_id=case_id
        ).model_dump()
        LOG.debug("Adding parameters from the pipeline config file if it exist")

        workflow_parameters: dict = built_workflow_parameters | (
            read_yaml(self.params) if hasattr(self, "params") and self.params else {}
        )
        replaced_workflow_parameters: dict = self.replace_values_in_params_file(
            workflow_parameters=workflow_parameters
        )
        if not dry_run:
            self.write_params_file(
                case_id=case_id, replaced_workflow_parameters=replaced_workflow_parameters
            )

    def replace_values_in_params_file(self, workflow_parameters: dict) -> dict:
        replaced_workflow_parameters = copy.deepcopy(workflow_parameters)
        """Iterate through the dictionary until all placeholders are replaced with the corresponding value from the dictionary"""
        while True:
            resolved: bool = True
            for key, value in replaced_workflow_parameters.items():
                new_value: str | int = self.replace_params_placeholders(value, workflow_parameters)
                if new_value != value:
                    resolved = False
                    replaced_workflow_parameters[key] = new_value
            if resolved:
                break
        return replaced_workflow_parameters

    def replace_params_placeholders(self, value: str | int, workflow_parameters: dict) -> str:
        """Replace values marked as placeholders with values from the given dictionary"""
        if isinstance(value, str):
            placeholders: list[str] = re.findall(r"{{\s*([^{}\s]+)\s*}}", value)
            for placeholder in placeholders:
                if placeholder in workflow_parameters:
                    value = value.replace(
                        f"{{{{{placeholder}}}}}", str(workflow_parameters[placeholder])
                    )
        return value

    def create_nextflow_config(self, case_id: str, dry_run: bool = False) -> None:
        """Create nextflow config file."""
        if content := self.get_nextflow_config_content(case_id=case_id):
            LOG.debug("Writing nextflow config file")
            if not dry_run:
                write_txt(
                    content=content,
                    file_path=self.get_nextflow_config_path(case_id=case_id),
                )

    def create_gene_panel(self, case_id: str, dry_run: bool) -> None:
        """Create and write an aggregated gene panel file exported from Scout."""
        LOG.info("Creating gene panel file")
        bed_lines: list[str] = self.get_gene_panel(case_id=case_id, dry_run=dry_run)
        if dry_run:
            bed_lines: str = "\n".join(bed_lines)
            LOG.debug(f"{bed_lines}")
            return
        self.write_panel(case_id=case_id, content=bed_lines)

    def config_case(self, case_id: str, dry_run: bool):
        """Create directory and config files required by a workflow for a case."""
        if dry_run:
            LOG.info("Dry run: Config files will not be written")
        self.status_db.verify_case_exists(case_internal_id=case_id)
        self.create_case_directory(case_id=case_id, dry_run=dry_run)
        self.create_sample_sheet(case_id=case_id, dry_run=dry_run)
        self.create_params_file(case_id=case_id, dry_run=dry_run)
        self.create_nextflow_config(case_id=case_id, dry_run=dry_run)
        if self.is_gene_panel_required:
            self.create_gene_panel(case_id=case_id, dry_run=dry_run)
        if self.is_managed_variants_required:
            vcf_lines: list[str] = self.get_managed_variants(case_id=case_id)
            if dry_run:
                for line in vcf_lines:
                    LOG.debug(line)
            else:
                self.write_managed_variants(case_id=case_id, content=vcf_lines)

    def _run_analysis_with_nextflow(
        self, case_id: str, command_args: NfCommandArgs, dry_run: bool
    ) -> None:
        """Run analysis with given options using Nextflow."""
        self.process = Process(
            binary=self.nextflow_binary_path,
            environment=self.conda_env,
            conda_binary=self.conda_binary,
            launch_directory=self.get_case_path(case_id=case_id),
        )
        LOG.info("Workflow will be executed using Nextflow")
        parameters: list[str] = NextflowHandler.get_nextflow_run_parameters(
            case_id=case_id,
            workflow_bin_path=self.workflow_bin_path,
            root_dir=self.root_dir,
            command_args=command_args.dict(),
        )
        self.process.export_variables(
            export=NextflowHandler.get_variables_to_export(),
        )
        command: str = self.process.get_command(parameters=parameters)
        LOG.info(f"{command}")
        sbatch_number: int = NextflowHandler.execute_head_job(
            case_id=case_id,
            case_directory=self.get_case_path(case_id=case_id),
            slurm_account=self.account,
            email=self.email,
            qos=self.get_slurm_qos_for_case(case_id=case_id),
            commands=command,
            dry_run=dry_run,
        )
        LOG.info(f"Nextflow head job running as job: {sbatch_number}")

    def _run_analysis_with_tower(
        self, case_id: str, command_args: NfCommandArgs, dry_run: bool
    ) -> str | None:
        """Run analysis with given options using NF-Tower."""
        LOG.info("Workflow will be executed using Tower")
        if command_args.resume:
            from_tower_id: int = command_args.id or NfTowerHandler.get_last_tower_id(
                case_id=case_id,
                trailblazer_config=self.get_job_ids_path(case_id=case_id),
            )
            LOG.info(f"Workflow will be resumed from run with Tower id: {from_tower_id}.")
            parameters: list[str] = NfTowerHandler.get_tower_relaunch_parameters(
                from_tower_id=from_tower_id, command_args=command_args.dict()
            )
        else:
            parameters: list[str] = NfTowerHandler.get_tower_launch_parameters(
                tower_workflow=self.tower_workflow, command_args=command_args.dict()
            )
        self.process.run_command(parameters=parameters, dry_run=dry_run)
        if self.process.stderr:
            LOG.error(self.process.stderr)
        if not dry_run:
            tower_id = NfTowerHandler.get_tower_id(stdout_lines=self.process.stdout_lines())
            self.write_trailblazer_config(case_id=case_id, tower_id=tower_id)
            return tower_id
        LOG.info(self.process.stdout)

    def get_command_args(
        self,
        case_id: str,
        work_dir: str,
        from_start: bool,
        profile: str,
        config: str,
        params_file: str | None,
        revision: str,
        compute_env: str,
        nf_tower_id: str | None,
        stub_run: bool,
    ) -> NfCommandArgs:
        command_args: NfCommandArgs = NfCommandArgs(
            **{
                "log": self.get_log_path(case_id=case_id, workflow=self.workflow),
                "work_dir": self.get_workdir_path(case_id=case_id, work_dir=work_dir),
                "resume": not from_start,
                "profile": self.get_profile(profile=profile),
                "config": self.get_nextflow_config_path(case_id=case_id, nextflow_config=config),
                "params_file": self.get_params_file_path(case_id=case_id, params_file=params_file),
                "name": case_id,
                "compute_env": compute_env or self.get_compute_env(case_id=case_id),
                "revision": revision or self.revision,
                "wait": NfTowerStatus.SUBMITTED,
                "id": nf_tower_id,
                "stub_run": stub_run,
            }
        )
        return command_args

    def run_nextflow_analysis(
        self,
        case_id: str,
        use_nextflow: bool,
        work_dir: str,
        from_start: bool,
        profile: str,
        config: str,
        params_file: str | None,
        revision: str,
        compute_env: str,
        stub_run: bool,
        nf_tower_id: str | None = None,
        dry_run: bool = False,
    ) -> None:
        """Prepare and start run analysis: check existence of all input files generated by config-case and sync with trailblazer."""
        self.status_db.verify_case_exists(case_internal_id=case_id)

        command_args = self.get_command_args(
            case_id=case_id,
            work_dir=work_dir,
            from_start=from_start,
            profile=profile,
            config=config,
            params_file=params_file,
            revision=revision,
            compute_env=compute_env,
            nf_tower_id=nf_tower_id,
            stub_run=stub_run,
        )

        try:
            self.verify_sample_sheet_exists(case_id=case_id, dry_run=dry_run)
            self.check_analysis_ongoing(case_id=case_id)
            LOG.info(f"Running analysis for {case_id}")
            tower_workflow_id: str | None = self.run_analysis(
                case_id=case_id,
                command_args=command_args,
                use_nextflow=use_nextflow,
                dry_run=dry_run,
            )
            self.set_statusdb_action(case_id=case_id, action=CaseActions.RUNNING, dry_run=dry_run)
        except FileNotFoundError as error:
            LOG.error(f"Could not resume analysis: {error}")
            raise FileNotFoundError
        except ValueError as error:
            LOG.error(f"Could not run analysis: {error}")
            raise ValueError
        except CgError as error:
            LOG.error(f"Could not run analysis: {error}")
            raise CgError

        if not dry_run:
            self.add_pending_trailblazer_analysis(
                case_id=case_id,
                tower_workflow_id=tower_workflow_id,
            )

    def run_analysis(
        self,
        case_id: str,
        command_args: NfCommandArgs,
        use_nextflow: bool,
        dry_run: bool = False,
    ) -> str | None:
        """Execute run analysis with given options."""
        if use_nextflow:
            self._run_analysis_with_nextflow(
                case_id=case_id,
                command_args=command_args,
                dry_run=dry_run,
            )
        else:
            return self._run_analysis_with_tower(
                case_id=case_id,
                command_args=command_args,
                dry_run=dry_run,
            )

    def get_deliverables_template_content(self) -> list[dict[str, str]]:
        """Return deliverables file template content."""
        LOG.debug("Getting deliverables file template content")
        return ReadFile.get_content_from_file(
            file_format=FileFormat.YAML,
            file_path=self.get_bundle_filenames_path(),
        )

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return bundle filenames path."""
        raise NotImplementedError

    @staticmethod
    def get_formatted_file_deliverable(
        file_template: dict[str | None, str | None],
        case_id: str,
        sample_id: str,
        sample_name: str,
        case_path: str,
    ) -> FileDeliverable:
        """Return the formatted file deliverable with the case and sample attributes."""
        deliverables = file_template.copy()
        for deliverable_field, deliverable_value in file_template.items():
            if deliverable_value is None:
                continue
            deliverables[deliverable_field] = (
                deliverables[deliverable_field]
                .replace("CASEID", case_id)
                .replace("SAMPLEID", sample_id)
                .replace("SAMPLENAME", sample_name)
                .replace("PATHTOCASE", case_path)
            )
        return FileDeliverable(**deliverables)

    def get_deliverables_for_sample(
        self, sample: Sample, case_id: str, template: list[dict[str, str]]
    ) -> list[FileDeliverable]:
        """Return a list of FileDeliverables for each sample."""
        sample_id: str = sample.internal_id
        sample_name: str = sample.name
        case_path = str(self.get_case_path(case_id=case_id))
        files: list[FileDeliverable] = []
        for file in template:
            files.append(
                self.get_formatted_file_deliverable(
                    file_template=file,
                    case_id=case_id,
                    sample_id=sample_id,
                    sample_name=sample_name,
                    case_path=case_path,
                )
            )
        return files

    def get_deliverables_for_case(self, case_id: str) -> WorkflowDeliverables:
        """Return workflow deliverables for a given case."""
        deliverable_template: list[dict] = self.get_deliverables_template_content()
        samples: list[Sample] = self.status_db.get_samples_by_case_id(case_id=case_id)
        files: list[FileDeliverable] = []

        for sample in samples:
            bundles_per_sample = self.get_deliverables_for_sample(
                sample=sample, case_id=case_id, template=deliverable_template
            )
            files.extend(bundle for bundle in bundles_per_sample if bundle not in files)
        return WorkflowDeliverables(files=files)

    def get_multiqc_json_path(self, case_id: str) -> Path:
        """Return the path of the multiqc_data.json file."""
        return Path(
            self.root_dir,
            case_id,
            MultiQC.MULTIQC,
            MultiQC.MULTIQC_DATA,
            MultiQC.MULTIQC_DATA + FileExtensions.JSON,
        )

    def get_workflow_metrics(self, metric_id: str) -> dict:
        """Get nf-core workflow metrics constants."""
        return {}

    def get_multiqc_search_patterns(self, case_id: str) -> dict:
        """Return search patterns for MultiQC. Each key is a search pattern and each value
        corresponds to the metric ID to set in the metrics deliverables file.
        Multiple search patterns can be added. Ideally, used patterns should be sample ids, e.g.
        {sample_id_1: sample_id_1, sample_id_2: sample_id_2}."""
        sample_ids: Iterator[str] = self.status_db.get_sample_ids_by_case_id(case_id)
        return {sample_id: sample_id for sample_id in sample_ids}

    @staticmethod
    def get_deduplicated_metrics(metrics: list[MetricsBase]) -> list[MetricsBase]:
        """Return deduplicated metrics based on metric ID and name. If duplicated entries are found
        only the first one will be kept."""
        deduplicated_metric_id_name = set([])
        deduplicated_metrics: list = []
        for metric in metrics:
            if (metric.id, metric.name) not in deduplicated_metric_id_name:
                deduplicated_metric_id_name.add((metric.id, metric.name))
                deduplicated_metrics.append(metric)
        return deduplicated_metrics

    def get_multiqc_data_json(self, case_id: str) -> MultiqcDataJson:
        return MultiqcDataJson(**read_json(file_path=self.get_multiqc_json_path(case_id=case_id)))

    def get_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the metrics specified in a MultiQC json file."""
        multiqc_json: MultiqcDataJson = self.get_multiqc_data_json(case_id=case_id)
        metrics = []
        for search_pattern, metric_id in self.get_multiqc_search_patterns(case_id=case_id).items():
            metrics_for_pattern: list[MetricsBase] = (
                self.get_metrics_from_multiqc_json_with_pattern(
                    search_pattern=search_pattern,
                    multiqc_json=multiqc_json,
                    metric_id=metric_id,
                    exact_match=self.is_multiqc_pattern_search_exact,
                )
            )
            metrics.extend(metrics_for_pattern)
        metrics = self.get_deduplicated_metrics(metrics=metrics)
        return metrics

    @staticmethod
    def _is_pattern_found(pattern: str, text: str, exact_match: bool) -> bool:
        if exact_match:
            is_pattern_found: bool = pattern == text
        else:
            is_pattern_found: bool = pattern in text
        return is_pattern_found

    def get_metrics_from_multiqc_json_with_pattern(
        self,
        search_pattern: str,
        multiqc_json: MultiqcDataJson,
        metric_id: str,
        exact_match: bool = False,
    ) -> list[MetricsBase]:
        """Parse a MultiqcDataJson and returns a list of metrics."""
        metrics: list[MetricsBase] = []
        for section in multiqc_json.report_general_stats_data:
            for subsection, metrics_dict in section.items():
                if self._is_pattern_found(
                    pattern=search_pattern, text=subsection, exact_match=exact_match
                ):
                    for metric_name, metric_value in metrics_dict.items():
                        metric: MetricsBase = self.get_multiqc_metric(
                            metric_name=metric_name, metric_value=metric_value, metric_id=metric_id
                        )
                        metrics.append(metric)
        return metrics

    def get_multiqc_metric(
        self, metric_name: str, metric_value: str | int | float, metric_id: str
    ) -> MetricsBase:
        """Return a MetricsBase object for a given metric."""
        return MetricsBase(
            header=None,
            id=metric_id,
            input=MultiQC.MULTIQC_DATA + FileExtensions.JSON,
            name=metric_name,
            step=MultiQC.MULTIQC,
            value=metric_value,
            condition=self.get_workflow_metrics(metric_id).get(metric_name, None),
        )

    @staticmethod
    def ensure_mandatory_metrics_present(metrics: list[MetricsBase]) -> None:
        return None

    def create_metrics_deliverables_content(self, case_id: str) -> dict[str, list[dict[str, Any]]]:
        """Create the content of metrics deliverables file."""
        metrics: list[MetricsBase] = self.get_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)
        return {"metrics": [metric.dict() for metric in metrics]}

    def write_metrics_deliverables(self, case_id: str, dry_run: bool = False) -> None:
        """Write <case>_metrics_deliverables.yaml file."""
        metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
        content: dict = self.create_metrics_deliverables_content(case_id=case_id)
        if dry_run:
            LOG.info(
                f"Dry-run: metrics deliverables file would be written to {metrics_deliverables_path.as_posix()}"
            )
            return

        LOG.info(f"Writing metrics deliverables file to {metrics_deliverables_path.as_posix()}")
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.YAML,
            file_path=metrics_deliverables_path,
        )

    def validate_qc_metrics(self, case_id: str, dry_run: bool = False) -> None:
        """Validate the information from a QC metrics deliverable file."""

        if dry_run:
            LOG.info("Dry-run: QC metrics validation would be performed")
            return

        LOG.info("Validating QC metrics")
        try:
            metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
            qc_metrics_raw: dict = ReadFile.get_content_from_file(
                file_format=FileFormat.YAML, file_path=metrics_deliverables_path
            )
            MetricsDeliverablesCondition(**qc_metrics_raw)
        except MetricsQCError as error:
            LOG.error(f"QC metrics failed for {case_id}, with: {error}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.FAILED)
            samples = self.status_db.get_samples_by_case_id(case_id=case_id)
            for sample in samples:
                trailblazer_comment = str(error).replace(
                    f"{sample.internal_id} - ", f"{sample.name} ({sample.internal_id}) - "
                )
            self.trailblazer_api.add_comment(case_id=case_id, comment=trailblazer_comment)
            raise MetricsQCError from error
        except CgError as error:
            LOG.error(f"Could not create metrics deliverables file: {error}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.ERROR)
            raise CgError from error
        self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.COMPLETED)

    def metrics_deliver(self, case_id: str, dry_run: bool):
        """Create and validate a metrics deliverables file for given case id."""
        self.status_db.verify_case_exists(case_internal_id=case_id)
        self.write_metrics_deliverables(case_id=case_id, dry_run=dry_run)
        self.validate_qc_metrics(case_id=case_id, dry_run=dry_run)

    def report_deliver(self, case_id: str, dry_run: bool = False, force: bool = False) -> None:
        """Write deliverables file."""
        self.status_db.verify_case_exists(case_id)
        self.trailblazer_api.verify_latest_analysis_is_completed(case_id=case_id, force=force)
        if dry_run:
            LOG.info(f"Dry-run: Would have created delivery files for case {case_id}")
            return
        workflow_content: WorkflowDeliverables = self.get_deliverables_for_case(case_id=case_id)
        self.write_deliverables_file(
            deliverables_content=workflow_content.dict(),
            file_path=self.get_deliverables_file_path(case_id=case_id),
        )
        LOG.info(
            f"Writing deliverables file in {self.get_deliverables_file_path(case_id=case_id).as_posix()}"
        )

    def store_analysis_housekeeper(
        self, case_id: str, comment: str | None = None, dry_run: bool = False, force: bool = False
    ) -> None:
        """
        Store a finished Nextflow analysis in Housekeeper and StatusDB.

        Raises:
            HousekeeperStoreError: If the deliverables file is malformed or if the bundle could not be stored.
        """
        try:
            self.status_db.verify_case_exists(case_id)
            self.trailblazer_api.verify_latest_analysis_is_completed(case_id=case_id, force=force)
            self.verify_deliverables_file_exists(case_id)
            self.upload_bundle_housekeeper(case_id=case_id, dry_run=dry_run, force=force)
            self.upload_bundle_statusdb(
                case_id=case_id, comment=comment, dry_run=dry_run, force=force
            )
            self.set_statusdb_action(case_id=case_id, action=None, dry_run=dry_run)
        except ValidationError as error:
            raise HousekeeperStoreError(f"Deliverables file is malformed: {error}")
        except Exception as error:
            self.housekeeper_api.rollback()
            self.status_db.session.rollback()
            raise HousekeeperStoreError(
                f"Could not store bundle in Housekeeper and StatusDB: {error}"
            )

    def store(
        self, case_id: str, comment: str | None = None, dry_run: bool = False, force: bool = False
    ):
        """Generate deliverable files for a case and store in Housekeeper if they
        pass QC metrics checks."""
        is_latest_analysis_qc: bool = self.trailblazer_api.is_latest_analysis_qc(case_id)
        is_latest_analysis_completed: bool = self.trailblazer_api.is_latest_analysis_completed(
            case_id
        )
        if not is_latest_analysis_qc and not is_latest_analysis_completed and not force:
            LOG.error(
                "Case not stored. Trailblazer status must be either QC or COMPLETE to be able to store"
            )
            raise ValueError

        if (
            is_latest_analysis_qc
            or not self.get_metrics_deliverables_path(case_id=case_id).exists()
            and not force
        ):
            LOG.info(f"Generating metrics file and performing QC checks for {case_id}")
            self.metrics_deliver(case_id=case_id, dry_run=dry_run)
        LOG.info(f"Storing analysis for {case_id}")
        self.report_deliver(case_id=case_id, dry_run=dry_run, force=force)
        self.store_analysis_housekeeper(
            case_id=case_id, comment=comment, dry_run=dry_run, force=force
        )

    def get_cases_to_store(self) -> list[Case]:
        """Return cases where analysis finished successfully,
        and is ready to be stored in Housekeeper."""
        return [
            case
            for case in self.status_db.get_running_cases_in_workflow(workflow=self.workflow)
            if self.trailblazer_api.is_latest_analysis_completed(case_id=case.internal_id)
            or self.trailblazer_api.is_latest_analysis_qc(case_id=case.internal_id)
        ]

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        """Return reference genome version for a case.
        Raises CgError if this information is missing or inconsistent for the samples linked to a case.
        """
        case = self.status_db.get_case_by_internal_id(case_id)
        return get_genome_build(case)

    def get_gene_panel_genome_build(self, case_id: str) -> GenePanelGenomeBuild:
        """Return build version of the gene panel for a case."""
        reference_genome: GenomeVersion = self.get_genome_build(case_id=case_id)
        try:
            return getattr(GenePanelGenomeBuild, reference_genome)
        except AttributeError as error:
            raise CgError(
                f"Reference {reference_genome} has no associated genome build for panels: {error}"
            ) from error

    def get_gene_panel(self, case_id: str, dry_run: bool = False) -> list[str]:
        """Create and return the aggregated gene panel file."""
        return self._get_gene_panel(
            case_id=case_id,
            genome_build=self.get_gene_panel_genome_build(case_id=case_id),
            dry_run=dry_run,
        )

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> NextflowAnalysis:
        """Parse Nextflow output analysis files and return an analysis model."""
        sample_metrics: dict[str, dict] = {}
        for metric in qc_metrics_raw:
            try:
                sample_metrics[metric.id].update({metric.name.lower(): metric.value})
            except KeyError:
                sample_metrics[metric.id] = {metric.name.lower(): metric.value}
        return NextflowAnalysis(sample_metrics=sample_metrics)

    def get_latest_metadata(self, case_id: str) -> NextflowAnalysis:
        """Return analysis output of a Nextflow case."""
        qc_metrics: list[MetricsBase] = self.get_multiqc_json_metrics(case_id)
        return self.parse_analysis(qc_metrics_raw=qc_metrics)

    def clean_past_run_dirs(self, before_date: str, skip_confirmation: bool = False) -> None:
        """Clean past run directories"""
        before_date: datetime = parse(before_date)
        analyses_to_clean: list[Analysis] = self.get_analyses_to_clean(before_date)
        LOG.info(f"Cleaning {len(analyses_to_clean)} analyses created before {before_date}")

        for analysis in analyses_to_clean:
            case_id = analysis.case.internal_id
            case_path = self.get_case_path(case_id)
            try:
                LOG.info(f"Cleaning output for {case_id}")
                self.clean_run_dir(
                    case_id=case_id, skip_confirmation=skip_confirmation, case_path=case_path
                )
            except FileNotFoundError:
                continue
            except Exception as error:
                LOG.error(f"Failed to clean directories for case {case_id} - {repr(error)}")
        LOG.info(f"Done cleaning {self.workflow} output")



from pydantic import BaseModel, BeforeValidator
from typing_extensions import Annotated

from cg.constants import NA_FIELD
from cg.models.delivery_report.metadata import (
    BalsamicTargetedSampleMetadataModel,
    BalsamicWGSSampleMetadataModel,
    MipDNASampleMetadataModel,
    RarediseaseSampleMetadataModel,
    RnafusionSampleMetadataModel,
    TaxprofilerSampleMetadataModel,
    TomteSampleMetadataModel,
)
from cg.models.delivery_report.validators import (
    get_boolean_as_string,
    get_date_as_string,
    get_delivered_files_as_file_names,
    get_prep_category_as_string,
    get_report_string,
    get_sex_as_string,
)


class ApplicationModel(BaseModel):
    """
    Analysis application attributes model

    Attributes:
        tag: application identifier; source: StatusDB/application/tag
        version: application version; source: LIMS/sample/application_version
        prep_category: library preparation category; source: StatusDB/application/prep_category
        description: analysis description; source: StatusDB/application/description
        limitations: application limitations; source: StatusDB/application/limitations
        accredited: if the sample associated process is accredited or not; source: StatusDB/application/is_accredited
        external: whether the app tag is external or not; source: StatusDB/application/is_external
    """

    tag: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    version: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    prep_category: Annotated[str, BeforeValidator(get_prep_category_as_string)] = NA_FIELD
    description: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    details: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    limitations: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    workflow_limitations: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    accredited: bool | None = None
    external: bool | None = None


class MethodsModel(BaseModel):
    """
    Model describing the methods used for preparation and sequencing of the samples

    Attributes:
        library_prep: library preparation method; source: LIMS/sample/prep_method
        sequencing: sequencing procedure; source: LIMS/sample/sequencing_method
    """

    library_prep: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    sequencing: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD


class TimestampModel(BaseModel):
    """
    Model describing the processing timestamps of a specific sample

    Atributes:
        ordered_at: order date; source: StatusDB/sample/ordered_at
        received_at: arrival date; source: StatusDB/sample/received_at
        prepared_at: library preparation date; source: StatusDB/sample/prepared_at
        reads_updated_at: sequencing date; source: StatusDB/sample/reads_updated_at
    """

    ordered_at: Annotated[str, BeforeValidator(get_date_as_string)] = NA_FIELD
    received_at: Annotated[str, BeforeValidator(get_date_as_string)] = NA_FIELD
    prepared_at: Annotated[str, BeforeValidator(get_date_as_string)] = NA_FIELD
    reads_updated_at: Annotated[str, BeforeValidator(get_date_as_string)] = NA_FIELD


class SampleModel(BaseModel):
    """
    Sample attributes model

    Attributes:
        name: sample name; source: StatusDB/sample/name
        id: sample internal ID; source: StatusDB/sample/internal_id
        ticket: ticket number; source: StatusDB/sample/ticket_number
        status: sample status provided by the customer; source: StatusDB/family-sample/status
        sex: sample sex provided by the customer; source: StatusDB/sample/sex
        source: sample type/source; source: LIMS/sample/source
        tumour: whether the sample is a tumour or normal one; source: StatusDB/sample/is_tumour
        application: analysis application model
        methods: sample processing methods model
        metadata: sample associated metrics and trending data model
        timestamps: processing timestamp attributes
        delivered_files: list of analysis sample files to be delivered
        delivered_fastq_files: list of fastq sample files to be delivered
    """

    name: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    id: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    ticket: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    status: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    sex: Annotated[str, BeforeValidator(get_sex_as_string)] = NA_FIELD
    source: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    tumour: Annotated[str, BeforeValidator(get_boolean_as_string)] = NA_FIELD
    application: ApplicationModel
    methods: MethodsModel
    metadata: (
        BalsamicTargetedSampleMetadataModel
        | BalsamicWGSSampleMetadataModel
        | MipDNASampleMetadataModel
        | RarediseaseSampleMetadataModel
        | RnafusionSampleMetadataModel
        | TaxprofilerSampleMetadataModel
        | TomteSampleMetadataModel
    )
    timestamps: TimestampModel
    delivered_files: Annotated[
        list[str] | str, BeforeValidator(get_delivered_files_as_file_names)
    ] = None
    delivered_fastq_files: Annotated[
        list[str] | str, BeforeValidator(get_delivered_files_as_file_names)
    ] = None



from pathlib import Path

import mock

from cg.services.deliver_files.file_fetcher.models import (
    DeliveryFiles,
    SampleFile,
    CaseFile,
    DeliveryMetaData,
)
from cg.services.deliver_files.file_formatter.destination.abstract import (
    DeliveryDestinationFormatter,
)
import pytest

from cg.services.deliver_files.file_formatter.destination.models import (
    FormattedFiles,
    FormattedFile,
)


@pytest.mark.parametrize(
    "formatter_service, formatted_case_files, formatted_sample_files, case_files, sample_files",
    [
        (
            "generic_delivery_file_formatter",
            "empty_case_files",
            "expected_formatted_analysis_sample_files",
            "empty_case_files",
            "expected_moved_analysis_sample_delivery_files",
        ),
        (
            "generic_delivery_file_formatter",
            "expected_formatted_analysis_case_files",
            "expected_formatted_analysis_sample_files",
            "expected_moved_analysis_case_delivery_files",
            "expected_moved_analysis_sample_delivery_files",
        ),
    ],
)
def test_reformat_files(
    formatter_service: DeliveryDestinationFormatter,
    formatted_case_files: list[FormattedFile],
    formatted_sample_files: list[FormattedFile],
    case_files: list[CaseFile],
    sample_files: list[SampleFile],
    tmp_path: Path,
    request,
):
    # GIVEN a delivery file formatter, mocked delivery files and expected formatted files
    formatter_service = request.getfixturevalue(formatter_service)
    formatted_case_files = request.getfixturevalue(formatted_case_files)
    formatted_sample_files = request.getfixturevalue(formatted_sample_files)
    case_files = request.getfixturevalue(case_files)
    sample_files = request.getfixturevalue(sample_files)

    delivery_data = DeliveryMetaData(
        case_id="some_case",
        customer_internal_id="cust_id",
        ticket_id="ticket_id",
        customer_ticket_inbox=Path(tmp_path, "cust_id", "inbox"),
    )
    mock_delivery_files = DeliveryFiles(
        delivery_data=delivery_data, case_files=case_files, sample_files=sample_files
    )
    files = []
    files.extend(formatted_sample_files)
    if case_files:
        files.extend(formatted_case_files)

    expected_formatted_files = FormattedFiles(files=files)
    with mock.patch(
        "cg.services.deliver_files.file_formatter.files.sample_service.SampleFileFormatter.format_files",
        return_value=formatted_sample_files,
    ), mock.patch(
        "cg.services.deliver_files.file_formatter.files.case_service.CaseFileFormatter.format_files",
        return_value=formatted_case_files,
    ):
        # WHEN reformatting the delivery files
        formatted_files: FormattedFiles = formatter_service.format_files(mock_delivery_files)

    # THEN the delivery files should be reformatted
    assert formatted_files == expected_formatted_files



import logging
from pathlib import Path

from cg.constants.constants import FileFormat
from cg.constants.pacbio import PacBioDirsAndFiles
from cg.services.decompression_service.decompressor import Decompressor
from cg.services.run_devices.abstract_classes import RunValidator
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.pacbio.run_file_manager.models import PacBioRunValidatorFiles
from cg.services.run_devices.pacbio.run_file_manager.run_file_manager import PacBioRunFileManager
from cg.services.validate_file_transfer_service.validate_file_transfer_service import (
    ValidateFileTransferService,
)

LOG = logging.getLogger(__name__)


class PacBioRunValidator(RunValidator):
    """
    PacBio run validator.
    Ensure that the post-processing of a pacbio run can start.
    """

    def __init__(
        self,
        decompressor: Decompressor,
        file_transfer_validator: ValidateFileTransferService,
        file_manager: PacBioRunFileManager,
    ):
        self.decompressor = decompressor
        self.file_transfer_validator = file_transfer_validator
        self.file_manager = file_manager

    def ensure_post_processing_can_start(self, run_data: PacBioRunData) -> None:
        """
        Ensure that a post-processing run can start.
        1. Check if all files are present listed in a manifest file.
        2. Decompresses the zipped reports.
        3. Touches a file to indicate that the run is validated
        4. Skips validation if the run is already validated
        """
        if self._is_validated(run_data.full_path):
            LOG.debug(f"Run for {run_data.full_path} is validated.")
            return
        paths_information: PacBioRunValidatorFiles = self.file_manager.get_run_validation_files(
            run_data
        )
        self.file_transfer_validator.validate_file_transfer(
            manifest_file=paths_information.manifest_file,
            source_dir=run_data.full_path,
            manifest_file_format=FileFormat.TXT,
        )
        self.decompressor.decompress(
            source_path=paths_information.decompression_target,
            destination_path=paths_information.decompression_destination,
        )
        self._touch_is_validated(run_data.full_path)
        LOG.debug(f"Run for {run_data.full_path} is validated.")

    @staticmethod
    def _is_validated(run_path: Path) -> bool:
        return Path(run_path, PacBioDirsAndFiles.RUN_IS_VALID).exists()

    @staticmethod
    def _touch_is_validated(run_path: Path) -> None:
        Path(run_path, PacBioDirsAndFiles.RUN_IS_VALID).touch()



from pydantic import BaseModel, Field

from cg.constants.metrics import DemuxMetricsColumnNames, QualityMetricsColumnNames


class SequencingQualityMetrics(BaseModel):
    """Model for the BCL Convert quality metrics."""

    lane: int = Field(..., alias=QualityMetricsColumnNames.LANE)
    sample_internal_id: str = Field(..., alias=QualityMetricsColumnNames.SAMPLE_INTERNAL_ID)
    mean_quality_score_q30: float = Field(
        ..., alias=QualityMetricsColumnNames.MEAN_QUALITY_SCORE_Q30
    )
    quality_score_sum: int = Field(..., alias=QualityMetricsColumnNames.QUALITY_SCORE_SUM)
    q30_bases_percent: float = Field(..., alias=QualityMetricsColumnNames.Q30_BASES_PERCENT)
    yield_: int = Field(..., alias=QualityMetricsColumnNames.YIELD)
    yield_q30: int = Field(..., alias=QualityMetricsColumnNames.YIELD_Q30)


class DemuxMetrics(BaseModel):
    """Model for the BCL Convert demultiplexing metrics."""

    lane: int = Field(..., alias=DemuxMetricsColumnNames.LANE)
    sample_internal_id: str = Field(..., alias=DemuxMetricsColumnNames.SAMPLE_INTERNAL_ID)
    read_pair_count: int = Field(..., alias=DemuxMetricsColumnNames.READ_PAIR_COUNT)



"""Helper functions for compress cli."""

import datetime as dt
import logging
import os
from math import ceil
from pathlib import Path
from typing import Iterator

from housekeeper.store.models import Bundle, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.compression import CRUNCHY_MIN_GB_PER_PROCESS, MAX_READS_PER_GB
from cg.constants.slurm import SlurmProcessing
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.meta.compress.files import get_spring_paths
from cg.store.models import Case
from cg.store.store import Store
from cg.utils.date import get_date_days_ago

LOG = logging.getLogger(__name__)


def get_cases_to_process(days_back: int, store: Store, case_id: str | None = None) -> list[Case]:
    """Return cases to process."""
    cases: list[Case] = []
    if case_id:
        case: Case = store.get_case_by_internal_id(case_id)
        if not case:
            LOG.warning(f"Could not find case {case_id}")
            return
        if case.is_compressible:
            cases.append(case)
    else:
        date_threshold: dt.datetime = get_date_days_ago(days_ago=days_back)
        cases: list[Case] = store.get_cases_to_compress(date_threshold=date_threshold)
    return cases


def get_fastq_individuals(store: Store, case_id: str = None) -> Iterator[str]:
    """Fetch individual ids from cases that are ready for SPRING compression"""
    case_obj = store.get_case_by_internal_id(internal_id=case_id)
    if not case_obj:
        LOG.error(f"Could not find case {case_id}")
        raise CaseNotFoundError("")

    for link_obj in case_obj.links:
        yield link_obj.sample.internal_id


def set_memory_according_to_reads(
    sample_id: str, sample_reads: int | None = None, sample_process_mem: int | None = None
) -> int | None:
    """Set SLURM sample process memory depending on the number of sample reads if sample_process_mem is not set."""
    if sample_process_mem:
        return sample_process_mem
    if not sample_reads:
        LOG.debug(f"No reads recorded for sample: {sample_id}")
        return
    sample_process_mem: int = ceil((sample_reads / MAX_READS_PER_GB))
    if sample_process_mem < CRUNCHY_MIN_GB_PER_PROCESS:
        return CRUNCHY_MIN_GB_PER_PROCESS
    if CRUNCHY_MIN_GB_PER_PROCESS <= sample_process_mem < SlurmProcessing.MAX_NODE_MEMORY:
        return sample_process_mem
    return SlurmProcessing.MAX_NODE_MEMORY


def update_compress_api(
    compress_api: CompressAPI, dry_run: bool, hours: int = None, mem: int = None, ntasks: int = None
) -> None:
    """Update parameters in Compress API."""

    compress_api.set_dry_run(dry_run=dry_run)
    if mem:
        LOG.debug(f"Set Crunchy API SLURM mem to {mem}")
        compress_api.crunchy_api.slurm_memory = mem
    if hours:
        LOG.debug(f"Set Crunchy API SLURM hours to {hours}")
        compress_api.crunchy_api.slurm_hours = hours
    if ntasks:
        LOG.debug(f"Set Crunchy API SLURM number of tasks to {ntasks}")
        compress_api.crunchy_api.slurm_number_tasks = ntasks


# Functions to fix problematic spring files


def get_versions(hk_api: HousekeeperAPI, bundle_name: str = None) -> Iterator[Version]:
    """Generates versions from hk bundles.

    If no bundle name is given generate latest version for every bundle.
    """
    if bundle_name:
        bundle: Bundle = hk_api.bundle(bundle_name)
        if not bundle:
            LOG.info(f"Could not find bundle {bundle_name}")
            return
        bundles: list[Bundle] = [bundle]
    else:
        bundles: list[Bundle] = hk_api.bundles()

    for bundle in bundles:
        LOG.debug(f"Check for versions in {bundle.name}")
        last_version: Version = hk_api.last_version(bundle.name)
        if not last_version:
            LOG.warning(f"No bundle found for {bundle.name} in Housekeeper")
            return
        yield last_version


def get_true_dir(dir_path: Path) -> Path | None:
    """Loop over the files in a directory, if any symlinks are found return the parent dir of the
    origin file."""
    # Check if there are any links to fastq files in the directory
    for fastq_path in dir_path.rglob("*"):
        # Check if there are fastq symlinks that points to the directory where the spring
        # path is located
        if fastq_path.is_symlink():
            return Path(os.readlink(fastq_path)).parent
    LOG.info("Could not find any symlinked files")
    return None


def compress_sample_fastqs_in_cases(
    compress_api: CompressAPI,
    cases: list[Case],
    dry_run: bool,
    number_of_conversions: int,
    hours: int = None,
    mem: int = None,
    ntasks: int = None,
) -> None:
    """Compress sample FASTQs for samples in cases."""
    case_conversion_count: int = 0
    individuals_conversion_count: int = 0
    for case in cases:
        case_converted = True
        if case_conversion_count >= number_of_conversions:
            break

        LOG.debug(f"Searching for FASTQ files in case {case.internal_id}")
        if not case.links:
            continue
        for case_link in case.links:
            sample_process_mem: int | None = set_memory_according_to_reads(
                sample_process_mem=mem,
                sample_id=case_link.sample.internal_id,
                sample_reads=case_link.sample.reads,
            )
            update_compress_api(
                compress_api=compress_api,
                dry_run=dry_run,
                hours=hours,
                mem=sample_process_mem,
                ntasks=ntasks,
            )
            case_converted: bool = compress_api.compress_fastq(
                sample_id=case_link.sample.internal_id
            )
            if not case_converted:
                LOG.debug(f"skipping individual {case_link.sample.internal_id}")
                continue
            individuals_conversion_count += 1
        if case_converted:
            case_conversion_count += 1
            LOG.info(f"Considering case {case.internal_id} converted")
    LOG.info(
        f"{individuals_conversion_count} individuals in {case_conversion_count} (completed) cases where compressed"
    )


def correct_spring_paths(
    hk_api: HousekeeperAPI, bundle_name: str = None, dry_run: bool = False
) -> None:
    """Function that will be used as a one off thing.

    There has been a problem when there are symlinked fastq files that are sent for compression.
    In these cases the spring archive has been created in the same place as that the symlinks are
    pointing to. This function will find those cases and move the spring archives to the correct
    place as specified in housekeeper.
    """
    versions = get_versions(hk_api=hk_api, bundle_name=bundle_name)
    for version_obj in versions:
        spring_paths = get_spring_paths(version_obj)
        i = 0
        for i, compression_obj in enumerate(spring_paths, 1):
            # We are interested in fixing the cases where spring paths are in wrong location
            spring_path = compression_obj.spring_path
            if spring_path.exists():
                continue

            # true_dir is where the spring paths actually exists
            true_dir = get_true_dir(spring_path.parent)
            if not true_dir:
                LOG.info("Could not find location of spring files")
                continue

            true_spring_path = true_dir / spring_path.name
            true_spring_config_path = true_spring_path.with_suffix("").with_suffix(".json")
            if not (true_spring_path.exists() and true_spring_config_path.exists()):
                LOG.info("Could not find spring and/or spring metadata files, skipping")
                continue
            LOG.info(
                f"Moving existing spring file (and config) {true_spring_path} to hk bundle path {spring_path}"
            )
            if not dry_run:
                # We know from above that the spring path does not exist
                true_spring_path.replace(spring_path)
                true_spring_config_path.replace(compression_obj.spring_metadata_path)
        if i == 0:
            LOG.debug("Could not find any spring files")



from cg.constants import Workflow
from cg.constants.delivery import PIPELINE_ANALYSIS_TAG_MAP
from cg.services.deliver_files.tag_fetcher.error_handling import (
    handle_tag_errors,
)
from cg.services.deliver_files.tag_fetcher.abstract import (
    FetchDeliveryFileTagsService,
)
from cg.services.deliver_files.tag_fetcher.models import DeliveryFileTags


class SampleAndCaseDeliveryTagsFetcher(FetchDeliveryFileTagsService):

    @handle_tag_errors
    def fetch_tags(self, workflow: Workflow) -> DeliveryFileTags:
        """Get the case tags for the files that need to be delivered for a workflow.
        args:
            workflow: The workflow to fetch tags for
        """
        self._validate_workflow(workflow)
        return DeliveryFileTags(
            case_tags=PIPELINE_ANALYSIS_TAG_MAP[workflow]["case_tags"],
            sample_tags=PIPELINE_ANALYSIS_TAG_MAP[workflow]["sample_tags"],
        )

    @staticmethod
    def _validate_workflow(workflow: Workflow):
        """Validate the workflow."""
        if workflow not in PIPELINE_ANALYSIS_TAG_MAP:
            raise ValueError(f"Workflow {workflow} is not supported.")
        return workflow



"""Test the cli for uploading using auto"""

import datetime
import logging
import pytest

from click.testing import CliRunner

from cg.cli.upload.base import upload_all_completed_analyses
from cg.constants import Workflow
from cg.models.cg_config import CGConfig
from tests.store_helpers import StoreHelpers

WORKFLOWS_TO_TEST: list = [
    Workflow.BALSAMIC,
    Workflow.MICROSALT,
    Workflow.MIP_DNA,
    Workflow.MIP_RNA,
    Workflow.RAREDISEASE,
    Workflow.RNAFUSION,
    Workflow.TAXPROFILER,
    Workflow.TOMTE,
]


@pytest.mark.parametrize(
    "workflow",
    WORKFLOWS_TO_TEST,
)
def test_upload_auto_with_workflow(
    cli_runner: CliRunner,
    workflow: Workflow,
    helpers: StoreHelpers,
    timestamp: datetime.datetime,
    upload_context: CGConfig,
    caplog,
):
    """Test upload auto: get analyses to upload, test upload completed."""
    # GIVEN a store with a workflow
    helpers.add_analysis(store=upload_context.status_db, completed_at=timestamp, workflow=workflow)

    # WHEN uploading all analyses
    caplog.set_level(logging.INFO)
    cli_runner.invoke(upload_all_completed_analyses, ["--workflow", workflow], obj=upload_context)
    # THEN assert that the MIP analysis was successfully uploaded
    assert "Uploading analysis for case" in caplog.text



"""Common CLI file generation functions"""

import rich_click as click

from cg.cli.generate.delivery_report.base import (
    generate_available_delivery_reports,
    generate_delivery_report,
)
from cg.cli.utils import CLICK_CONTEXT_SETTINGS


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
def generate():
    """Generates and/or modifies files."""


generate.add_command(generate_delivery_report)
generate.add_command(generate_available_delivery_reports)



import logging

import mock
import pytest
from click.testing import CliRunner, Result

from cg.cli.upload.scout import create_scout_load_config, get_upload_api
from cg.constants import EXIT_SUCCESS, Workflow
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.meta.workflow.tomte import TomteAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from cg.store.store import Store
from tests.mocks.scout import MockScoutLoadConfig
from tests.store_helpers import StoreHelpers

WORKFLOWS_TO_TEST: list = [
    Workflow.BALSAMIC,
    Workflow.MIP_DNA,
    Workflow.MIP_RNA,
    Workflow.RNAFUSION,
]

ANALYSIS_API: list = [
    (Workflow.BALSAMIC, BalsamicAnalysisAPI),
    (Workflow.MIP_DNA, MipDNAAnalysisAPI),
    (Workflow.MIP_RNA, MipRNAAnalysisAPI),
    (Workflow.RNAFUSION, RnafusionAnalysisAPI),
    (Workflow.TOMTE, TomteAnalysisAPI),
]


@pytest.mark.parametrize(
    "workflow,analysis_api",
    ANALYSIS_API,
)
def test_get_upload_api(
    cg_context: CGConfig,
    case_id: str,
    helpers: StoreHelpers,
    workflow: Workflow,
    analysis_api: AnalysisAPI,
):
    """Test to get the correct upload API for a case."""
    status_db: Store = cg_context.status_db

    # GIVEN a case with a balsamic analysis
    case: Case = helpers.ensure_case(store=status_db, data_analysis=workflow, case_id=case_id)
    helpers.add_analysis(store=status_db, case=case, workflow=workflow)

    # WHEN getting the upload API
    upload_api: UploadAPI = get_upload_api(cg_config=cg_context, case=case)

    # THEN assert that the type of upload API is correct
    assert isinstance(upload_api.analysis_api, analysis_api)


@pytest.mark.parametrize(
    "workflow",
    WORKFLOWS_TO_TEST,
)
def test_create_scout_load_config(
    caplog,
    cli_runner: CliRunner,
    cg_context: CGConfig,
    case_id: str,
    helpers: StoreHelpers,
    workflow: Workflow,
):
    """Test to create a scout load config for a case."""
    caplog.set_level(logging.DEBUG)
    status_db: Store = cg_context.status_db

    # GIVEN a case with a balsamic analysis
    case: Case = helpers.ensure_case(store=status_db, data_analysis=workflow, case_id=case_id)
    helpers.add_analysis(store=status_db, case=case, workflow=workflow)

    with mock.patch.object(UploadScoutAPI, "generate_config", return_value=MockScoutLoadConfig()):
        # WHEN creating the scout load config
        result: Result = cli_runner.invoke(
            create_scout_load_config, ["--print", case_id], obj=cg_context
        )

        # THEN assert that the code exits with success
        assert result.exit_code == EXIT_SUCCESS
        # THEN assert that the root path is in the log
        assert cg_context.meta_apis["upload_api"].analysis_api.root in caplog.text



import pytest

from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.services.orders.store_order_services.store_case_order import StoreCaseOrderService
from cg.services.orders.store_order_services.store_fastq_order_service import StoreFastqOrderService
from cg.services.orders.store_order_services.store_metagenome_order import (
    StoreMetagenomeOrderService,
)
from cg.services.orders.store_order_services.store_microbial_fastq_order_service import (
    StoreMicrobialFastqOrderService,
)
from cg.services.orders.store_order_services.store_microbial_order import StoreMicrobialOrderService
from cg.services.orders.store_order_services.store_pacbio_order_service import (
    StorePacBioOrderService,
)
from cg.services.orders.store_order_services.store_pool_order import StorePoolOrderService


@pytest.fixture
def balsamic_status_data(
    balsamic_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
) -> dict:
    """Parse balsamic order example."""
    project: OrderType = OrderType.BALSAMIC
    order: OrderIn = OrderIn.parse_obj(obj=balsamic_order_to_submit, project=project)
    return store_generic_order_service.order_to_status(order=order)


@pytest.fixture
def fastq_status_data(
    fastq_order_to_submit, store_fastq_order_service: StoreFastqOrderService
) -> dict:
    """Parse fastq order example."""
    project: OrderType = OrderType.FASTQ
    order: OrderIn = OrderIn.parse_obj(obj=fastq_order_to_submit, project=project)
    return store_fastq_order_service.order_to_status(order=order)


@pytest.fixture
def pacbio_status_data(
    pacbio_order_to_submit: dict, store_pacbio_order_service: StorePacBioOrderService
) -> dict:
    """Parse pacbio order example."""
    project: OrderType = OrderType.PACBIO_LONG_READ
    order: OrderIn = OrderIn.parse_obj(obj=pacbio_order_to_submit, project=project)
    return store_pacbio_order_service.order_to_status(order=order)


@pytest.fixture
def metagenome_status_data(
    metagenome_order_to_submit: dict, store_metagenome_order_service: StoreMetagenomeOrderService
) -> dict:
    """Parse metagenome order example."""
    project: OrderType = OrderType.METAGENOME
    order: OrderIn = OrderIn.parse_obj(obj=metagenome_order_to_submit, project=project)

    return store_metagenome_order_service.order_to_status(order=order)


@pytest.fixture
def microbial_status_data(
    microbial_order_to_submit: dict, store_microbial_order_service: StoreMicrobialOrderService
) -> dict:
    """Parse microbial order example."""
    project: OrderType = OrderType.MICROSALT
    order: OrderIn = OrderIn.parse_obj(obj=microbial_order_to_submit, project=project)
    return store_microbial_order_service.order_to_status(order=order)


@pytest.fixture
def microbial_fastq_status_data(
    microbial_fastq_order_to_submit: dict,
    store_microbial_fastq_order_service: StoreMicrobialFastqOrderService,
) -> dict:
    """Parse microbial order example."""
    project: OrderType = OrderType.MICROBIAL_FASTQ
    order: OrderIn = OrderIn.parse_obj(obj=microbial_fastq_order_to_submit, project=project)
    return store_microbial_fastq_order_service.order_to_status(order=order)


@pytest.fixture
def mip_rna_status_data(
    mip_rna_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
) -> dict:
    """Parse rna order example."""
    project: OrderType = OrderType.MIP_RNA
    order: OrderIn = OrderIn.parse_obj(obj=mip_rna_order_to_submit, project=project)
    return store_generic_order_service.order_to_status(order=order)


@pytest.fixture
def mip_status_data(
    mip_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
) -> dict:
    """Parse scout order example."""
    project: OrderType = OrderType.MIP_DNA
    order: OrderIn = OrderIn.parse_obj(obj=mip_order_to_submit, project=project)
    return store_generic_order_service.order_to_status(order=order)


@pytest.fixture
def rml_status_data(
    rml_order_to_submit: dict, store_pool_order_service: StorePoolOrderService
) -> dict:
    """Parse rml order example."""
    project: OrderType = OrderType.RML
    order: OrderIn = OrderIn.parse_obj(obj=rml_order_to_submit, project=project)
    return store_pool_order_service.order_to_status(order=order)


@pytest.fixture
def tomte_status_data(
    tomte_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
) -> dict:
    """Parse TOMTE order example."""
    project: OrderType = OrderType.TOMTE
    order: OrderIn = OrderIn.parse_obj(obj=tomte_order_to_submit, project=project)
    return store_generic_order_service.order_to_status(order=order)



"""Tests for the PacBioRunDataGenerator."""

from pathlib import Path

import pytest

from cg.services.run_devices.exc import PostProcessingRunDataGeneratorError
from cg.services.run_devices.pacbio.run_data_generator.pacbio_run_data_generator import (
    PacBioRunDataGenerator,
)
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData


@pytest.mark.parametrize(
    "run_name_fixture, smrt_cell_name_fixture, run_data_fixture",
    [
        (
            "pac_bio_test_run_name",
            "pac_bio_smrt_cell_name",
            "expected_pac_bio_run_data_1_b01",
        ),
        ("pacbio_barcoded_run_name", "pacbio_barcoded_smrt_cell_name", "pacbio_barcoded_run_data"),
    ],
    ids=["1_B01", "1_C01"],
)
def test_get_run_data(
    pac_bio_run_data_generator: PacBioRunDataGenerator,
    pac_bio_runs_dir: Path,
    run_name_fixture: str,
    smrt_cell_name_fixture: str,
    run_data_fixture: str,
    request: pytest.FixtureRequest,
):
    # GIVEN a run directory, a run name and a SMRT cell name
    run_name: str = request.getfixturevalue(run_name_fixture)
    smrt_cell_name: str = request.getfixturevalue(smrt_cell_name_fixture)
    expected_run_data: PacBioRunData = request.getfixturevalue(run_data_fixture)
    run_name: str = "/".join([run_name, smrt_cell_name])

    # WHEN Generating run data
    run_data: PacBioRunData = pac_bio_run_data_generator.get_run_data(
        run_name=run_name, sequencing_dir=pac_bio_runs_dir.as_posix()
    )

    # THEN the correct run data are returned
    assert run_data == expected_run_data


@pytest.mark.parametrize("wrong_run_name", ["rimproper_name", "d_improper_name "])
def test_get_run_data_improper_name(
    pac_bio_run_data_generator: PacBioRunDataGenerator,
    pac_bio_runs_dir: Path,
    wrong_run_name: str,
):
    # GIVEN a PacBioRunDataGenerator and a wrong run name

    # WHEN Generating run data with the wrong run name

    # THEN an PostProcessingRunDataGeneratorError is raised
    with pytest.raises(PostProcessingRunDataGeneratorError):
        pac_bio_run_data_generator.get_run_data(
            run_name=wrong_run_name, sequencing_dir=pac_bio_runs_dir.as_posix()
        )



import logging
import os
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import SampleFile, CaseFile

LOG = logging.getLogger(__name__)


class FileManager:
    """
    Service to manage files.
    Handles operations that create or rename files and directories.
    """

    @staticmethod
    def create_directories(base_path: Path, directories: set[str]) -> None:
        """Create directories for given names under the base path.
        args:
            base_path: The base path to create the directories under.
            directories: The directories to create within the given base path. Can be a list of one.
        """

        for directory in directories:
            LOG.debug(f"[FileManager] Creating directory or file: {base_path}/{directory}")
            Path(base_path, directory).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def rename_file(src: Path, dst: Path) -> None:
        """
        Rename a file from src to dst.
        raise ValueError if src does not exist.
        args:
            src: The source file path.
            dst: The destination file path.
        """
        if not src or not dst:
            raise ValueError("Source and destination paths cannot be None.")
        LOG.debug(f"[FileManager] Renaming file: {src} -> {dst}")
        if not src.exists():
            raise FileNotFoundError(f"Source file {src} does not exist.")
        os.rename(src=src, dst=dst)

    @staticmethod
    def create_hard_link(src: Path, dst: Path) -> None:
        """
        Create a hard link from src to dst.
        args:
            src: The source file path.
            dst: The destination file path.
        """
        LOG.debug(f"[FileManager] Creating hard link: {src} -> {dst}")
        os.link(src=src, dst=dst)


class FileMover:
    """
    Service class to move files.
    Requires a file management service to perform file operations.
    """

    def __init__(self, file_manager):
        """
        args:
         file_manager: Service for file operations (e.g., create directories, move files).
        """
        self.file_management_service = file_manager

    def create_directories(self, base_path: Path, directories: set[str]) -> None:
        """Create required directories.
        args:
            base_path: The base path to create the directories under.
            directories: The directories to create.
        """
        self.file_management_service.create_directories(base_path, directories)

    def move_files_to_directory(self, file_models: list, target_dir: Path) -> None:
        """Move files to the target directory.
        args:
            file_models: The file models that contain the files to move.
            target_dir: The directory to move the files to.
        """
        for file_model in file_models:
            target_path = Path(target_dir, file_model.file_path.name)
            self._move_or_link_file(src=file_model.file_path, dst=target_path)

    @staticmethod
    def update_file_paths(
        file_models: list[CaseFile | SampleFile], target_dir: Path
    ) -> list[CaseFile | SampleFile]:
        """Update file paths to point to the target directory.
        args:
            file_models: The file models to update.
            target_dir: The target directory to point the file paths to.
        """
        for file_model in file_models:
            file_model.file_path = Path(target_dir, file_model.file_path.name)
        return file_models

    def move_and_update_files(
        self, file_models: list[CaseFile | SampleFile], target_dir: Path
    ) -> list[CaseFile | SampleFile]:
        """Move files to the target directory and update the file paths.
        args:
            file_models: The file models that contain the files to move.
            target_dir: The directory to move the files to.
        """
        if file_models:
            self.move_files_to_directory(file_models=file_models, target_dir=target_dir)
            return self.update_file_paths(file_models=file_models, target_dir=target_dir)
        return file_models

    def _move_or_link_file(self, src: Path, dst: Path) -> None:
        """Move or create a hard link for a file.
        args:
            src: The source file path
            dst: The destination file path
        """
        LOG.debug(f"[FileMover] Moving file: {src} -> {dst}")
        if dst.exists():
            LOG.debug(f"Overwriting existing file: {dst}")
            dst.unlink()
        self.file_management_service.create_hard_link(src=src, dst=dst)



from pathlib import Path

import pytest
from pydantic import ValidationError

from cg.exc import NfSampleSheetError
from cg.models.rnafusion.rnafusion import RnafusionSampleSheetEntry


def test_instantiate_rnafusion_sample(
    sample_name: str,
    fastq_forward_read_path: Path,
    fastq_reverse_read_path: Path,
    strandedness: str,
):
    """
    Tests Rnafusion sample with correct values.
    """
    # GIVEN a sample with fastq files and strandedness

    # WHEN instantiating a rnafusion sample object
    rnafusion_sample_object = RnafusionSampleSheetEntry(
        name=sample_name,
        fastq_forward_read_paths=[fastq_forward_read_path, fastq_forward_read_path],
        fastq_reverse_read_paths=[fastq_reverse_read_path, fastq_reverse_read_path],
        strandedness=strandedness,
    )

    # THEN assert that it was successfully created
    assert isinstance(rnafusion_sample_object, RnafusionSampleSheetEntry)


def test_incomplete_fastq_file_pairs(
    sample_name: str,
    fastq_forward_read_path: Path,
    fastq_reverse_read_path: Path,
    strandedness: str,
):
    """
    Tests Rnafusion sample with lists of different length for fastq forward and reverse reads.
    """
    # GIVEN a sample with fastq files of different lengths

    # WHEN instantiating a sample object

    # THEN throws an error
    with pytest.raises(NfSampleSheetError) as error:
        RnafusionSampleSheetEntry(
            name=sample_name,
            fastq_forward_read_paths=[fastq_forward_read_path, fastq_forward_read_path],
            fastq_reverse_read_paths=[fastq_reverse_read_path],
            strandedness=strandedness,
        )
    assert "Fastq file length for forward and reverse do not match" in str(error.value)


def test_fastq_empty_list(
    sample_name: str,
    fastq_forward_read_path: Path,
    fastq_reverse_read_path: Path,
    strandedness: str,
):
    """
    Tests Rnafusion sample with an empty list for fastq reverse reads (single end).
    """

    # GIVEN a sample with fastq files and strandedness

    # WHEN instantiating a sample object

    # THEN throws an error
    with pytest.raises(ValidationError):
        RnafusionSampleSheetEntry(
            name=sample_name,
            fastq_forward_read_paths=[fastq_forward_read_path, fastq_forward_read_path],
            fastq_reverse_read_paths=[],
            strandedness=strandedness,
        )
        # THEN errors should be raised


def test_strandedness_not_permitted(
    sample_name: str,
    fastq_forward_read_path: Path,
    fastq_reverse_read_path: Path,
    strandedness_not_permitted: str,
):
    """
    Tests Rnafusion sample with not permitted strandedness.
    """

    # WHEN instantiating a sample object THEN throws a ValidationError
    with pytest.raises(ValidationError):
        RnafusionSampleSheetEntry(
            name=sample_name,
            fastq_forward_read_paths=[fastq_forward_read_path],
            fastq_reverse_read_paths=[fastq_reverse_read_path],
            strandedness=strandedness_not_permitted,
        )
        # THEN errors should be raised


def test_non_existing_fastq_file(
    sample_name: str,
    fastq_forward_read_path: Path,
    non_existing_file_path: Path,
    strandedness: str,
):
    """
    Tests Rnafusion sample with non-existing files
    """
    # GIVEN a sample with a non-existing file

    # WHEN instantiating a sample object THEN throws an error
    with pytest.raises(NfSampleSheetError) as error:
        RnafusionSampleSheetEntry(
            name=sample_name,
            fastq_forward_read_paths=[fastq_forward_read_path],
            fastq_reverse_read_paths=[non_existing_file_path],
            strandedness=strandedness,
        )
    assert "Fastq file does not exist" in str(error.value)



"""Fixtures for the tests of the IlluminaPostProcessingService."""

from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.devices import DeviceType
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.post_processing.post_processing_service import (
    IlluminaPostProcessingService,
)
from cg.store.models import IlluminaFlowCell
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def illumina_post_postprocessing_service(
    store: Store,
    real_housekeeper_api: HousekeeperAPI,
    selected_novaseq_x_sample_ids: list[str],
    helpers: StoreHelpers,
    illumina_demultiplexed_runs_post_processing_hk_api: HousekeeperAPI,
    tmp_illumina_demultiplexed_runs_directory: str,
) -> IlluminaPostProcessingService:
    """Return an instance of the IlluminaPostProcessingService."""
    for sample_id in selected_novaseq_x_sample_ids:
        helpers.add_sample(store, internal_id=sample_id)
    return IlluminaPostProcessingService(
        status_db=store,
        housekeeper_api=illumina_demultiplexed_runs_post_processing_hk_api,
        dry_run=False,
        demultiplexed_runs_dir=Path(tmp_illumina_demultiplexed_runs_directory),
    )


@pytest.fixture
def illumina_flow_cell(
    novaseq_x_demux_runs_flow_cell: IlluminaRunDirectoryData,
) -> IlluminaFlowCell:
    """Return an Illumina flow cell."""
    return IlluminaFlowCell(
        internal_id=novaseq_x_demux_runs_flow_cell.id,
        type=DeviceType.ILLUMINA,
        model=novaseq_x_demux_runs_flow_cell.run_parameters.get_flow_cell_model(),
    )



from cg.exc import OrderError
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import PacBioSample
from cg.services.orders.submitters.order_submitter import ValidateOrderService
from cg.store.models import ApplicationVersion, Customer
from cg.store.store import Store


class ValidatePacbioOrderService(ValidateOrderService):

    def __init__(self, status_db: Store):
        self.status_db = status_db

    def validate_order(self, order: OrderIn) -> None:
        self._validate_customer_exists(order.customer)
        self._validate_applications_exist(order.samples)
        self._validate_sample_names_available(samples=order.samples, customer_id=order.customer)

    def _validate_customer_exists(self, customer_id: str) -> None:
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        if not customer:
            raise OrderError(f"Unknown customer: {customer_id}")

    def _validate_applications_exist(self, samples: list[PacBioSample]) -> None:
        for sample in samples:
            application_tag = sample.application
            application_version: ApplicationVersion = (
                self.status_db.get_current_application_version_by_tag(tag=application_tag)
            )
            if application_version is None:
                raise OrderError(f"Invalid application: {sample.application}")

    def _validate_sample_names_available(
        self, samples: list[PacBioSample], customer_id: str
    ) -> None:
        customer: Customer = self.status_db.get_customer_by_internal_id(customer_id)
        for sample in samples:
            if self.status_db.get_sample_by_customer_and_name(
                customer_entry_id=[customer.id], sample_name=sample.name
            ):
                raise OrderError(
                    f"Sample name already used in a previous order by the same customer: {sample.name}"
                )



import pytest

from cg.models.orders.constants import OrderType
from cg.store.exc import EntryNotFoundError
from cg.store.models import Application
from cg.store.store import Store


def test_get_active_applications_by_order_type_no_application(store: Store):
    """Test that if there are not applications for a given type an error is raised."""
    # GIVEN a store with applications without order types
    applications: list[Application] = store.get_applications()
    for application in applications:
        assert not application.order_type_applications

    # GIVEN an order type
    order_type: OrderType = OrderType.PACBIO_LONG_READ

    # WHEN getting active applications by order type
    with pytest.raises(EntryNotFoundError):
        store.get_active_applications_by_order_type(order_type)

    # THEN an EntryNotFoundError is raised



"""Tests for the mapping utils module."""

import pytest

from cg.exc import CgError
from cg.utils.mapping import get_item_by_pattern_in_source


def test_get_item_by_pattern():
    # GIVEN a mapping dictionary

    mapping_dict: dict[str, any] = {"key_one": "one", "key_two": "two", "key_three": "three"}

    # WHEN retrieving an item by a pattern
    item: any = get_item_by_pattern_in_source(source="key_three_dee", pattern_map=mapping_dict)

    # THEN the correct items is returned
    assert item == "three"


def test_get_item_by_pattern_raises_error():
    # GIVEN a mapping dictionary

    mapping_dict: dict[str, any] = {"key_one": "one", "key_two": "two", "key_three": "three"}

    # WHEN retrieving an item by a pattern
    with pytest.raises(CgError):
        get_item_by_pattern_in_source(source="key_not_there", pattern_map=mapping_dict)

    # THEN an error is raised



"""Fixtures for lims tests"""

import pytest

from cg.apps.lims.api import LimsAPI
from tests.mocks.limsmock import MockReagentType


class MockLims(LimsAPI):
    """Mock parts of the lims api"""

    lims = None

    def __init__(self):
        """Mock the init method"""
        self.lims = self

    def get_prepmethod(self, lims_id: str) -> str:
        """Override the get_prepmethod"""

    def get_sequencingmethod(self, lims_id: str) -> str:
        """Override the get_sequencingmethod"""

    def get_deliverymethod(self, lims_id: str) -> str:
        """Override the get_deliverymethod"""


@pytest.fixture(scope="function")
def lims_mock():
    """Returns a Lims api mock"""
    _lims_api = MockLims()
    return _lims_api


@pytest.fixture
def reagent() -> MockReagentType:
    """Return a valid MockReagentType."""
    return MockReagentType(label="A702-A506 (ACAGTGGT-CTAGAACA)", sequence="ACAGTGGT-CTAGAACA")


@pytest.fixture
def reagent_label(reagent: MockReagentType) -> str:
    """Returns a reagent label."""
    return reagent.label


@pytest.fixture
def reagent_sequence(reagent: MockReagentType) -> str:
    """Returns a reagent sequence."""
    return reagent.sequence


@pytest.fixture
def invalid_reagent() -> MockReagentType:
    """Return a MockReagentType with invalid data."""
    return MockReagentType(label="A702-A506 (ACAGTGGT-CTAGAACA)", sequence="CTAGAACA-ACAGTGGT")


@pytest.fixture
def invalid_reagent_label(invalid_reagent: MockReagentType) -> str:
    """Returns a reagent label."""
    return invalid_reagent.label


@pytest.fixture
def invalid_reagent_sequence(invalid_reagent: MockReagentType) -> str:
    """Returns a reagent sequence."""
    return invalid_reagent.sequence


@pytest.fixture
def label_no_parentheses() -> str:
    """Returns a reagent label."""
    return "ACAGTGGT-CTAGAACA"



"""Module to test the StoreGenericOrderService class."""

import datetime as dt
import math
from copy import deepcopy

from cg.constants import DataDelivery, Priority, Workflow
from cg.models.orders.order import OrderIn, OrderType
from cg.services.orders.store_order_services.store_case_order import (
    StoreCaseOrderService,
)
from cg.store.models import Sample
from cg.store.store import Store


def test_cases_to_status(
    mip_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
):
    # GIVEN a scout order with a trio case
    project: OrderType = OrderType.MIP_DNA
    order = OrderIn.parse_obj(mip_order_to_submit, project=project)

    # WHEN parsing for status
    data = store_generic_order_service.order_to_status(order=order)

    # THEN it should pick out the case
    assert len(data["families"]) == 2
    family = data["families"][0]
    assert family["name"] == "family1"
    assert family["data_analysis"] == Workflow.MIP_DNA
    assert family["data_delivery"] == str(DataDelivery.SCOUT)
    assert family["priority"] == Priority.standard.name
    assert family["cohorts"] == ["Other"]
    assert (
        family["synopsis"]
        == "As for the synopsis it will be this overly complex sentence to prove that the synopsis field might in fact be a very long string, which we should be prepared for."
    )
    assert set(family["panels"]) == {"IEM"}
    assert len(family["samples"]) == 3

    first_sample = family["samples"][0]
    assert math.isclose(first_sample["age_at_sampling"], 17.18192, rel_tol=1e-09, abs_tol=1e-09)
    assert first_sample["name"] == "sample1"
    assert first_sample["application"] == "WGSPCFC030"
    assert first_sample["phenotype_groups"] == ["Phenotype-group"]
    assert first_sample["phenotype_terms"] == ["HP:0012747", "HP:0025049"]
    assert first_sample["sex"] == "female"
    assert first_sample["status"] == "affected"
    assert first_sample["subject_id"] == "subject1"
    assert first_sample["mother"] == "sample2"
    assert first_sample["father"] == "sample3"

    # ... second sample has a comment
    assert isinstance(family["samples"][1]["comment"], str)


def test_cases_to_status_synopsis(
    mip_order_to_submit: dict, store_generic_order_service: StoreCaseOrderService
):
    # GIVEN a scout order with a trio case where synopsis is None
    modified_order: dict = deepcopy(mip_order_to_submit)
    for sample in modified_order["samples"]:
        sample["synopsis"] = None

    project: OrderType = OrderType.MIP_DNA
    order = OrderIn.parse_obj(mip_order_to_submit, project=project)

    # WHEN parsing for status
    store_generic_order_service.order_to_status(order=order)

    # THEN No exception should have been raised on synopsis


def test_store_mip(
    base_store: Store,
    mip_status_data: dict,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):
    # GIVEN a basic store with no samples or nothing in it + scout order
    assert not base_store._get_query(table=Sample).first()
    assert not base_store.get_cases()

    # WHEN storing the order
    new_families = store_generic_order_service.store_items_in_status(
        customer_id=mip_status_data["customer"],
        order=mip_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_status_data["families"],
    )

    # THEN it should create and link samples and the case
    assert len(new_families) == 2
    new_case = new_families[0]
    assert new_case.name == "family1"
    assert set(new_case.panels) == {"IEM"}
    assert new_case.priority_human == Priority.standard.name

    assert len(new_case.links) == 3
    new_link = new_case.links[0]
    assert new_case.data_analysis == Workflow.MIP_DNA
    assert new_case.data_delivery == str(DataDelivery.SCOUT)
    assert set(new_case.cohorts) == {"Other"}
    assert (
        new_case.synopsis
        == "As for the synopsis it will be this overly complex sentence to prove that the synopsis field might in fact be a very long string, which we should be prepared for."
    )
    assert new_link.status == "affected"
    assert new_link.mother.name == "sample2"
    assert new_link.father.name == "sample3"
    assert new_link.sample.name == "sample1"
    assert new_link.sample.sex == "female"
    assert new_link.sample.application_version.application.tag == "WGSPCFC030"
    assert new_link.sample.is_tumour
    assert isinstance(new_case.links[1].sample.comment, str)

    assert set(new_link.sample.phenotype_groups) == {"Phenotype-group"}
    assert set(new_link.sample.phenotype_terms) == {"HP:0012747", "HP:0025049"}
    assert new_link.sample.subject_id == "subject1"
    assert math.isclose(new_link.sample.age_at_sampling, 17.18192, rel_tol=1e-09, abs_tol=1e-09)


def test_store_mip_rna(
    base_store: Store,
    mip_rna_status_data,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):
    # GIVEN a basic store with no samples or nothing in it + rna order
    rna_application_tag = "RNAPOAR025"
    assert not base_store._get_query(table=Sample).first()
    assert not base_store.get_cases()
    assert base_store.get_application_by_tag(tag=rna_application_tag)

    # WHEN storing the order
    new_cases = store_generic_order_service.store_items_in_status(
        customer_id=mip_rna_status_data["customer"],
        order=mip_rna_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=mip_rna_status_data["families"],
    )

    # THEN it should create and link samples and the casing
    assert len(new_cases) == 1
    new_casing = new_cases[0]

    assert len(new_casing.links) == 2
    new_link = new_casing.links[0]
    assert new_casing.data_analysis == Workflow.MIP_RNA
    assert new_casing.data_delivery == str(DataDelivery.SCOUT)
    assert new_link.sample.name == "sample1-rna-t1"
    assert new_link.sample.application_version.application.tag == rna_application_tag


def test_store_cancer_samples(
    base_store: Store,
    balsamic_status_data: dict,
    ticket_id: str,
    store_generic_order_service: StoreCaseOrderService,
):

    # GIVEN a basic store with no samples and a cancer order
    assert not base_store._get_query(table=Sample).first()
    assert not base_store.get_cases()

    # WHEN storing the order
    new_families = store_generic_order_service.store_items_in_status(
        customer_id=balsamic_status_data["customer"],
        order=balsamic_status_data["order"],
        ordered=dt.datetime.now(),
        ticket_id=ticket_id,
        items=balsamic_status_data["families"],
    )

    # THEN it should create and link samples and the case
    assert len(new_families) == 1
    new_case = new_families[0]
    assert new_case.name == "family1"
    assert new_case.data_analysis in [
        Workflow.BALSAMIC,
        Workflow.BALSAMIC_QC,
        Workflow.BALSAMIC_UMI,
    ]
    assert new_case.data_delivery == str(DataDelivery.FASTQ_ANALYSIS_SCOUT)
    assert not set(new_case.panels)
    assert new_case.priority_human == Priority.standard.name

    assert len(new_case.links) == 1
    new_link = new_case.links[0]
    assert new_link.sample.name == "s1"
    assert new_link.sample.sex == "male"
    assert new_link.sample.application_version.application.tag == "WGSPCFC030"
    assert new_link.sample.comment == "other Elution buffer"
    assert new_link.sample.is_tumour



import pytest

from cg.cli.post_process.utils import (
    UnprocessedRunInfo,
    get_post_processing_service_from_run_name,
    get_unprocessed_runs_info,
)
from cg.models.cg_config import CGConfig
from cg.services.run_devices.pacbio.post_processing_service import PacBioPostProcessingService


def test_get_post_processing_service_from_run_name(
    pac_bio_context: CGConfig, pacbio_barcoded_sequencing_run_name: str
):
    """Test that a the correct post processing service is returned given a run name."""
    # GIVEN a context with a post-processing service for the run name
    assert pac_bio_context.post_processing_services.pacbio

    # WHEN getting the post-processing service from the run name
    service: PacBioPostProcessingService = get_post_processing_service_from_run_name(
        context=pac_bio_context, run_name=pacbio_barcoded_sequencing_run_name
    )
    # THEN a the correct post-processing service should be returned
    assert isinstance(service, PacBioPostProcessingService)


@pytest.mark.parametrize(
    "wrong_run_name",
    [
        "m_84202_2024_05_22_133539/1_A01",
        "r84202_20240522133539/1_A01",
        "r84202_20240522_133539/3_A01",
    ],
    ids=["doesn't start with r", "incorrect splits", "incorrect plate"],
)
def test_get_post_processing_service_from_wrong_run_name(
    pac_bio_context: CGConfig, wrong_run_name: str
):
    """Test that a the correct post processing service is returned given a run name."""
    # GIVEN a wrong run name

    # GIVEN a context with a post-processing service for the run name
    assert pac_bio_context.post_processing_services.pacbio

    # WHEN getting the post-processing service from the run name
    with pytest.raises(NameError):
        get_post_processing_service_from_run_name(context=pac_bio_context, run_name=wrong_run_name)

    # THEN an error is raised


def test_get_unprocessed_runs_info_pacbio(
    pac_bio_context: CGConfig,
    pacbio_barcoded_sequencing_run_name: str,
    pacbio_sequencing_run_name: str,
):
    """Test that a list of unprocessed runs is returned for Pacbio."""
    # GIVEN a context with a post-processing service for Pacbio
    assert pac_bio_context.post_processing_services.pacbio
    instrument: str = "pacbio"

    # GIVEN that there are unprocessed run in the directory
    expected_run_names: set[str] = {pacbio_sequencing_run_name, pacbio_barcoded_sequencing_run_name}

    # GIVEN that there is an already processed run in the directory
    number_of_runs: int = len(pac_bio_context.run_names_services.pacbio.get_run_names())
    number_unprocessed: int = len(expected_run_names)
    assert number_of_runs > number_unprocessed

    # WHEN getting the unprocessed runs
    unprocessed_runs: list[UnprocessedRunInfo] = get_unprocessed_runs_info(
        context=pac_bio_context, instrument=instrument
    )

    # THEN the expected unprocessed runs are returned
    assert unprocessed_runs
    unprocessed_run_names: set[str] = {run.name for run in unprocessed_runs}
    assert unprocessed_run_names == expected_run_names



"""Tests for helper functions in Cg Compress CLI."""

import logging
from pathlib import Path
from typing import Iterator

from housekeeper.store.models import Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.compress import helpers
from cg.cli.compress.helpers import set_memory_according_to_reads
from cg.constants.compression import CRUNCHY_MIN_GB_PER_PROCESS, MAX_READS_PER_GB


def test_set_memory_according_to_reads_when_no_reads(caplog, sample_id: str):
    """Test setting memory according to reads when no sample reads."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a sample id and no reads supplied

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(sample_id=sample_id, sample_reads=0)

    # THEN we should log
    assert f"No reads recorded for sample: {sample_id}" in caplog.text

    # THEN None should be returned
    assert memory is None


def test_set_memory_according_to_reads_when_few_reads(sample_id: str):
    """Test setting memory according to reads when few reads."""
    # GIVEN a sample id and reads

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(sample_id=sample_id, sample_reads=1)

    # THEN memory should be set to the minimum
    assert memory == CRUNCHY_MIN_GB_PER_PROCESS


def test_set_memory_according_to_reads_when_many_reads(sample_id: str):
    """Test setting memory according to reads when many reads."""
    # GIVEN a sample id and reads

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(
        sample_id=sample_id, sample_reads=MAX_READS_PER_GB**10
    )

    # THEN memory should be limited to what is available on the node
    assert memory == 180


def test_set_memory_according_to_reads(sample_id: str):
    """Test setting memory according to reads."""
    # GIVEN a sample id and reads

    # WHEN setting memory according to reads
    memory: int = set_memory_according_to_reads(
        sample_id=sample_id, sample_reads=MAX_READS_PER_GB * 100
    )

    # THEN memory should be adjusted
    assert memory == 100


def test_get_true_dir_no_symlinks(project_dir: Path):
    """Test to get a true dir when there are no symlinks."""
    # GIVEN a directory with some files but no symlinked files
    a_file: Path = Path(project_dir, "hello.txt")
    a_file.touch()
    assert a_file.exists()

    # WHEN fetching the true dir for the files in fixture dir
    true_dir = helpers.get_true_dir(a_file.parent)

    # THEN assert that the true_dir is None since there where no symbolic links in the project_dir
    assert true_dir is None


def test_get_true_dir_when_symlinks(project_dir: Path):
    """Test to get a true dir when there are symlinks to another directory.

    The function should return the path to another directory when there are symlinks in one.
    """
    # GIVEN a directory with a file with some content
    content = "hello world"
    a_file: Path = Path(project_dir, "hello.txt")
    a_file.write_text(content)
    assert a_file.exists()

    # GIVEN a subdirectory with a link to the above file
    new_dir: Path = Path(project_dir, "new_dir/")
    new_dir.mkdir()
    a_link: Path = Path(new_dir, "link_to_hello")
    a_link.symlink_to(a_file)
    assert a_link.read_text() == content
    assert a_link.parent == new_dir
    assert a_link.is_symlink() is True

    # WHEN fetching the true dir for the symlinked dir
    true_dir = helpers.get_true_dir(a_link.parent)

    # THEN assert that the true_dir the same as the parent of the destination of the symlink
    assert true_dir == project_dir


def test_get_versions_no_bundle(housekeeper_api: HousekeeperAPI):
    """Test to get latest versions of bundles when no bundles exists."""
    # GIVEN a empty housekeeper_api

    # WHEN fetching versions
    versions: Iterator[Version] = helpers.get_versions(housekeeper_api)

    # THEN assert no versions was returned
    assert sum(1 for version in versions) == 0


def test_get_versions_one_bundle(housekeeper_api: HousekeeperAPI, spring_bundle: dict):
    """Test to get latest versions of bundles when no bundles exists."""
    # GIVEN a populated housekeeper_api
    housekeeper_api.add_bundle(spring_bundle)

    # WHEN fetching versions
    versions = helpers.get_versions(housekeeper_api)

    # THEN assert no versions was returned
    assert sum(1 for version in versions) == 1


def test_correct_spring_paths(
    housekeeper_api: HousekeeperAPI,
    spring_bundle_symlink_problem: dict,
    symlinked_fastqs: dict,
):
    """Test for correct_spring_paths."""
    # GIVEN a populated housekeeper_api
    housekeeper_api.add_bundle(spring_bundle_symlink_problem)
    # GIVEN that the spring files exists in the wrong location
    versions: Iterator[Version] = helpers.get_versions(housekeeper_api)
    version: Version = next(versions)
    for file_path in version.files:
        assert not Path(file_path.full_path).exists()

    # WHEN updating the spring paths
    helpers.correct_spring_paths(housekeeper_api)

    # THEN assert that the spring paths has been moved
    for file_path in version.files:
        assert Path(file_path.full_path).exists()



"""Tests for the read illumina flow cell functions in the store."""

from cg.store.models import IlluminaFlowCell
from cg.store.store import Store


def test_get_illumina_flow_cell_by_internal_id(
    illumina_flow_cell: IlluminaFlowCell, illumina_flow_cell_internal_id: str, store: Store
):
    # GIVEN an Illumina flow cell in store
    store.add_illumina_flow_cell(illumina_flow_cell)

    # WHEN getting the flow cell from the store
    flow_cell: IlluminaFlowCell = store.get_illumina_flow_cell_by_internal_id(
        illumina_flow_cell_internal_id
    )

    # THEN the correct flow cell should be returned
    assert isinstance(flow_cell, IlluminaFlowCell)
    assert flow_cell.internal_id == illumina_flow_cell_internal_id



from unittest import mock
from unittest.mock import Mock
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.file_fetcher.exc import NoDeliveryFilesError
from cg.services.deliver_files.file_fetcher.models import DeliveryFiles


def test_file_delivery_service_no_files(empty_delivery_files: DeliveryFiles):
    # GIVEN a delivery file service with no files to deliver
    file_delivery_service = DeliverFilesService(
        delivery_file_manager_service=Mock(),
        move_file_service=Mock(),
        file_formatter_service=Mock(),
        rsync_service=Mock(),
        tb_service=Mock(),
        analysis_service=Mock(),
        status_db=Mock(),
    )

    # WHEN getting the files to deliver that raises an NoDeliveryFilesError

    # THEN the service should return immediately without any errors
    with mock.patch.object(
        file_delivery_service.file_manager,
        "get_files_to_deliver",
        side_effect=NoDeliveryFilesError,
    ):
        file_delivery_service.deliver_files_for_case(case=Mock(), delivery_base_path=Mock())

    assert not file_delivery_service.file_mover.move_files.called
    assert not file_delivery_service.file_formatter.format_files.called
    assert not file_delivery_service.rsync_service.rsync_files.called
    assert not file_delivery_service.tb_service.add_rsync_job.called
    assert not file_delivery_service.analysis_service.update_status.called



from http import HTTPStatus

import pytest
from _pytest.fixtures import FixtureRequest
from pytest_mock import MockFixture

from cg.clients.janus.api import JanusAPIClient
from cg.clients.janus.dto.create_qc_metrics_request import (
    CreateQCMetricsRequest,
    FilePathAndTag,
    WorkflowInfo,
)
from cg.constants import Workflow
from cg.constants.sequencing import SeqLibraryPrepCategory


@pytest.fixture
def janus_config() -> dict:
    return {"janus": {"host": "janus_host"}}


@pytest.fixture
def janus_client(janus_config: dict) -> JanusAPIClient:
    return JanusAPIClient(janus_config)


@pytest.fixture
def balsamic_files_wgs(request: FixtureRequest) -> list[FilePathAndTag]:
    file_path_tags: dict = {
        "alignment_summary_metrics_path": "picard_alignment_summary_tag",
        "picard_hs_metrics_path": "picard_hs_metrics_tag",
        "picard_dups_path": "picard_dups_tag",
        "picard_wgs_metrics_path": "picard_wgs_metrics_tag",
        "picard_insert_size_path": "picard_insert_size_tag",
        "somalier_path": "somalier_tag",
        "fastp_path": "fastp_tag",
        "samtools_stats_path": "samtools_stats_tag",
    }

    files: list[FilePathAndTag] = []
    for key, value in file_path_tags.items():
        files.append(FilePathAndTag(file_path=str(key), tag=value))
    return files


@pytest.fixture
def collect_qc_request_balsamic_wgs(
    balsamic_files_wgs: list[FilePathAndTag],
) -> CreateQCMetricsRequest:
    workflow_info = WorkflowInfo(workflow=Workflow.BALSAMIC, version="1")
    return CreateQCMetricsRequest(
        case_id="testcase",
        sample_ids=["test_sample"],
        files=balsamic_files_wgs,
        workflow_info=workflow_info,
        prep_category=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
    )


@pytest.fixture
def janus_response() -> dict:
    return {"some": "data"}


@pytest.fixture
def mock_post_request_ok(janus_response: dict, mocker: MockFixture) -> MockFixture:
    mocked_response = mocker.Mock()
    mocked_response.status_code = HTTPStatus.OK
    mocked_response.json.return_value = janus_response
    mocked_response.verify = False
    return mocked_response


@pytest.fixture
def mock_post_request_not_found(janus_response: dict, mocker: MockFixture) -> MockFixture:
    mocked_response = mocker.Mock()
    mocked_response.status_code = HTTPStatus.NOT_FOUND
    mocked_response.json.return_value = janus_response
    return mocked_response



from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import DeliveryFiles, SampleFile, CaseFile
from cg.services.deliver_files.file_mover.abstract import DestinationFilesMover
from cg.services.deliver_files.utils import FileMover


class BaseDestinationFilesMover(DestinationFilesMover):
    """
    Class to move files directly to the delivery base path.
    """

    def __init__(self, file_mover: FileMover):
        self.file_mover = file_mover

    def move_files(self, delivery_files: DeliveryFiles, delivery_base_path: Path) -> DeliveryFiles:
        """
        Move the files directly to the delivery base path.
        args:
            delivery_files: DeliveryFiles: The files to move.
            delivery_base_path: Path: The path to move the files to.
        """
        delivery_files.delivery_data.delivery_path = delivery_base_path
        delivery_files.case_files = self.file_mover.move_and_update_files(
            file_models=delivery_files.case_files, target_dir=delivery_base_path
        )
        delivery_files.sample_files = self.file_mover.move_and_update_files(
            file_models=delivery_files.sample_files, target_dir=delivery_base_path
        )
        return delivery_files



"""Fixtures for crunchy API."""

import logging
from pathlib import Path

import pytest

from cg.apps.crunchy.models import CrunchyMetadata
from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile
from cg.models import CompressionData

LOG = logging.getLogger(__name__)


@pytest.fixture
def real_spring_metadata_path(apps_dir: Path) -> Path:
    """Return the path to a SPRING metadata file."""
    return Path(apps_dir, "crunchy", "spring_metadata.json")


@pytest.fixture
def spring_metadata(compression_object: CompressionData) -> list[dict]:
    """Return meta data information."""
    return [
        {
            "path": compression_object.fastq_first.as_posix(),
            "file": "first_read",
            "checksum": "checksum_first_read",
            "algorithm": "sha256",
        },
        {
            "path": compression_object.fastq_second.as_posix(),
            "file": "second_read",
            "checksum": "checksum_second_read",
            "algorithm": "sha256",
        },
        {"path": compression_object.spring_path.as_posix(), "file": "spring"},
    ]


@pytest.fixture
def crunchy_metadata_object(spring_metadata: list[dict]) -> CrunchyMetadata:
    """Return Crunchy metadata."""
    return CrunchyMetadata(files=spring_metadata)


@pytest.fixture
def spring_metadata_file(compression_object: CompressionData, spring_metadata: list[dict]) -> Path:
    """Return the path to a populated SPRING metadata file."""
    metadata_path = compression_object.spring_metadata_path
    WriteFile.write_file_from_content(
        content=spring_metadata, file_format=FileFormat.JSON, file_path=metadata_path
    )
    return metadata_path



import logging
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import SampleFile
from cg.services.deliver_files.file_formatter.files.abstract import FileFormatter
from cg.services.deliver_files.file_formatter.destination.models import FormattedFile
from cg.services.deliver_files.file_formatter.path_name.abstract import PathNameFormatter
from cg.services.deliver_files.utils import FileManager

LOG = logging.getLogger(__name__)


class SampleFileFormatter(FileFormatter):
    """
    Format the sample files to deliver.
    Used for all workflows except Microsalt and Mutant.
    args:
        file_manager: The file manager
        path_name_formatter: The path name formatter to format paths to either a flat or nested structure in the delivery destination
    """

    def __init__(
        self,
        file_manager: FileManager,
        path_name_formatter: PathNameFormatter,
    ):
        self.file_manager = file_manager
        self.path_name_formatter = path_name_formatter

    def format_files(
        self, moved_files: list[SampleFile], delivery_path: Path
    ) -> list[FormattedFile]:
        """
        Format the sample files to deliver and return the formatted files.
        args:
            moved_sample_files: The sample files to format. These are files that have been moved from housekeeper to the delivery path.
            delivery_path: The path to deliver the files to
        """
        LOG.debug("[FORMAT SERVICE] Formatting sample files")
        sample_names: set[str] = self._get_sample_names(sample_files=moved_files)
        for sample_name in sample_names:
            self.file_manager.create_directories(base_path=delivery_path, directories={sample_name})
        formatted_files: list[FormattedFile] = self._format_sample_file_paths(moved_files)
        for formatted_file in formatted_files:
            self.file_manager.rename_file(
                src=formatted_file.original_path, dst=formatted_file.formatted_path
            )
        return formatted_files

    @staticmethod
    def _get_sample_names(sample_files: list[SampleFile]) -> set[str]:
        """Extract sample names from the sample files."""
        return {sample_file.sample_name for sample_file in sample_files}

    def _format_sample_file_paths(self, sample_files: list[SampleFile]) -> list[FormattedFile]:
        """
        Return a list of formatted sample files.
        args:
            sample_files: The sample files to format
        """
        return [
            FormattedFile(
                original_path=sample_file.file_path,
                formatted_path=self.path_name_formatter.format_file_path(
                    file_path=sample_file.file_path,
                    provided_id=sample_file.sample_id,
                    provided_name=sample_file.sample_name,
                ),
            )
            for sample_file in sample_files
        ]



"""PacBio run data object fixtures."""

from pathlib import Path

import pytest

from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData


@pytest.fixture
def expected_pac_bio_run_data_1_b01(
    pac_bio_test_run_name: str, pac_bio_smrt_cell_dir_1_b01: Path
) -> PacBioRunData:
    return PacBioRunData(
        full_path=pac_bio_smrt_cell_dir_1_b01,
        sequencing_run_name=pac_bio_test_run_name,
        well_name="B01",
        plate=1,
    )


@pytest.fixture
def pacbio_barcoded_run_data(
    pacbio_barcoded_run_name: str, pacbio_barcoded_smrt_cell_dir_1_c01: Path
) -> PacBioRunData:
    return PacBioRunData(
        full_path=pacbio_barcoded_smrt_cell_dir_1_c01,
        sequencing_run_name=pacbio_barcoded_run_name,
        well_name="C01",
        plate=1,
    )



import pytest

from cg.services.illumina.backup.backup_service import IlluminaBackupService
from cg.meta.encryption.encryption import EncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import CGConfig
from cg.store.store import Store


@pytest.fixture
def backup_context(cg_context: CGConfig, store_with_illumina_sequencing_data: Store) -> CGConfig:
    cg_context.meta_apis["backup_api"] = IlluminaBackupService(
        encryption_api=EncryptionAPI(binary_path=cg_context.encryption.binary_path),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=store_with_illumina_sequencing_data,
        tar_api=TarAPI(binary_path=cg_context.tar.binary_path),
        pdc_service=cg_context.pdc_service,
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )
    cg_context.status_db_ = store_with_illumina_sequencing_data
    return cg_context



from datetime import datetime
from pathlib import Path

from pydantic import BaseModel, Field, ValidationInfo, field_validator

from cg.constants.constants import SampleType


class BalsamicConfigAnalysis(BaseModel):
    """Balsamic analysis model

    Attributes:
        case_id: case internal ID
        analysis_type: analysis type (single, paired or pon)
        analysis_workflow: analysis carried out (balsamic, balsamic-qc or balsamic-umi)
        sequencing_type: analysis sequencing type (wgs or targeted)
        BALSAMIC_version: BALSAMIC version used to produce the analysis
        config_creation_date: config creation timestamp
    """

    case_id: str
    analysis_type: str
    analysis_workflow: str
    sequencing_type: str
    BALSAMIC_version: str
    config_creation_date: datetime | str


class BalsamicConfigSample(BaseModel):
    """Sample attributes used for Balsamic analysis.

    Attributes:
        type (str): sample type (tumor or normal)
    """

    type: SampleType
    name: str
    fastq_info: dict[str, dict[str, Path]]


class BalsamicConfigReference(BaseModel):
    """Metadata of reference files.

    Attributes:
        reference_genome: reference genome fasta file
        reference_genome_version: reference genome build version
    """

    reference_genome: Path
    reference_genome_version: str | None = Field(default=None, validate_default=True)

    @field_validator("reference_genome_version")
    @classmethod
    def extract_genome_version_from_path(cls, _, info: ValidationInfo) -> str:
        """
        Return the genome version from the reference path:
        /home/proj/stage/cancer/balsamic_cache/X.X.X/hg19/genome/human_g1k_v37.fasta
        """

        return str(info.data.get("reference_genome")).split("/")[-3]


class BalsamicConfigPanel(BaseModel):
    """Balsamic attributes of a panel BED file.

    Attributes:
        capture_kit: string representation of a panel BED filename
        capture_kit_version: capture kit version
        chrom: list of chromosomes in the panel BED file
        pon_cnn: panel of normal name and version applied to the CNVkit variant calling
    """

    capture_kit: str
    capture_kit_version: str | None = Field(default=None, validate_default=True)
    chrom: list[str]
    pon_cnn: str | None = None

    @field_validator("capture_kit", mode="before")
    @classmethod
    def get_filename_from_path(cls, path: str) -> str:
        """Return the base name of the provided file path."""
        return Path(path).name

    @field_validator("capture_kit_version")
    @classmethod
    def get_panel_version_from_filename(cls, _, info: ValidationInfo) -> str:
        """Return the panel bed version from its filename (e.g. gicfdna_3.1_hg19_design.bed)."""
        return info.data.get("capture_kit").split("_")[-3]

    @field_validator("pon_cnn", mode="before")
    @classmethod
    def get_pon_cnn_name_version_from_filename(cls, pon_cnn: str | None) -> str:
        """Return the CNVkit PON name and version from its filename (gmsmyeloid_5.3_hg19_design_CNVkit_PON_reference_v1.cnn)."""
        pon_cnn_filename_split: list[str] = Path(pon_cnn).stem.split("_")
        pon_cnn_name: str = f"{pon_cnn_filename_split[0]} v{pon_cnn_filename_split[1]}"
        pon_cnn_version: str = pon_cnn_filename_split[-1]
        pon_tool_name: str = pon_cnn_filename_split[4]
        return f"{pon_tool_name} {pon_cnn_name} ({pon_cnn_version})"


class BalsamicConfigQC(BaseModel):
    """Config QC attributes.

    Attributes:
        adapter: adapter sequence that has been trimmed
        min_seq_length: minimum sequence length cutoff for reads
    """

    adapter: str | None
    min_seq_length: str | None


class BalsamicVarCaller(BaseModel):
    """Variant caller attributes model.

    Attributes:
        mutation: if it addresses somatic or germline mutations
        type: called variant type (CNV, SV, SNV, etc.)
        analysis_type: analysis type supported by the variant caller (single, paired or pon)
        sequencing_type: sequencing type supported (wgs or targeted)
        workflow_solution: supported workflow (BALSAMIC, Sentieon, Sentieon_umi, etc.)
    """

    mutation: str
    mutation_type: str
    analysis_type: list[str]
    sequencing_type: list[str]
    workflow_solution: list[str]


class BalsamicConfigJSON(BaseModel):
    """Model for BALSAMIC analysis config validation.

    Attributes:
        analysis: config analysis attributes
        samples: sample attributes associated to a specific case
        reference: BALSAMIC build reference
        panel: panel attributes (targeted analysis exclusively)
    """

    analysis: BalsamicConfigAnalysis
    samples: list[BalsamicConfigSample]
    reference: BalsamicConfigReference
    panel: BalsamicConfigPanel | None = None
    QC: BalsamicConfigQC
    vcf: dict[str, BalsamicVarCaller]
    bioinfo_tools_version: dict[str, list[str]]



"""Module for the collect qc metrics meta api."""

import logging
from typing import Iterator

from housekeeper.store.models import File, Tag

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.clients.arnold.api import ArnoldAPIClient
from cg.clients.arnold.dto.create_case_request import CreateCaseRequest
from cg.clients.janus.api import JanusAPIClient
from cg.clients.janus.dto.create_qc_metrics_request import (
    CreateQCMetricsRequest,
    FilePathAndTag,
    WorkflowInfo,
)
from cg.constants.housekeeper_tags import JanusTags
from cg.exc import HousekeeperFileMissingError
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class CollectQCMetricsAPI:
    """Collect qc metrics API."""

    def __init__(
        self,
        hk_api: HousekeeperAPI,
        status_db: Store,
        janus_api: JanusAPIClient,
        arnold_api: ArnoldAPIClient,
    ):
        self.hk_api: HousekeeperAPI = hk_api
        self.status_db: Store = status_db
        self.janus_api: JanusAPIClient = janus_api
        self.arnold_api: ArnoldAPIClient = arnold_api

    def get_qc_metrics_file_paths_for_case(self, case_id: str) -> list[File]:
        """Get the multiqc metrics files for a case."""
        files: list[File] = self.hk_api.get_files(
            bundle=case_id, tags=JanusTags.tags_to_retrieve
        ).all()
        if not files:
            raise HousekeeperFileMissingError(f"No qc metrics files found for {case_id}.")
        return files

    @staticmethod
    def get_tag_to_include(tags: list[Tag]) -> str:
        """Get the multiqc module file tag."""
        for tag in tags:
            if tag.name in JanusTags.multi_qc_file_tags:
                return tag.name

    def get_file_paths_and_tags(self, files: list[File]) -> list[FilePathAndTag]:
        """Get the file path and tag for the multiqc files."""
        file_paths_and_tags: list[FilePathAndTag] = []
        for hk_file in files:
            file_path: str = hk_file.full_path
            tag: str = self.get_tag_to_include(hk_file.tags)
            file_paths_and_tags.append(FilePathAndTag(file_path=file_path, tag=tag))
        return file_paths_and_tags

    @staticmethod
    def get_workflow_info(case: Case) -> WorkflowInfo:
        workflow: str = case.data_analysis
        workflow_version: str = case.analyses[0].workflow_version
        return WorkflowInfo(workflow=workflow, version=workflow_version)

    def create_qc_metrics_request(self, case_id: str) -> CreateQCMetricsRequest:
        """Create a qc metrics request."""
        files: list[File] = self.get_qc_metrics_file_paths_for_case(case_id)
        file_paths_and_tags: list[FilePathAndTag] = self.get_file_paths_and_tags(files)
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        workflow_info: WorkflowInfo = self.get_workflow_info(case)
        sample_ids: Iterator[str] = self.status_db.get_sample_ids_by_case_id(case_id)

        return CreateQCMetricsRequest(
            case_id=case_id,
            sample_ids=sample_ids,
            workflow_info=workflow_info,
            files=file_paths_and_tags,
        )

    def get_case_qc_metrics(self, case_id: str) -> dict:
        """Get the qc metrics for a case."""
        qc_metrics_request: CreateQCMetricsRequest = self.create_qc_metrics_request(case_id)
        qc_metrics: dict = self.janus_api.qc_metrics(qc_metrics_request)
        return qc_metrics

    def get_create_case_request(self, case_id: str) -> CreateCaseRequest:
        case_qc_metrics: dict = self.get_case_qc_metrics(case_id)
        if not case_qc_metrics:
            raise ValueError(f"Could not retrieve qc metrics from Janus for case {case_id}")
        return CreateCaseRequest(**case_qc_metrics)

    def create_case(self, case_id: str, dry_run: bool = False):
        case_request: CreateCaseRequest = self.get_create_case_request(case_id)
        if dry_run:
            LOG.info(f"Would have sent create case request to Arnold with: {case_request}")
            return
        self.arnold_api.create_case(case_request)



import shutil
from pathlib import Path

import pytest


@pytest.fixture
def zip_base_path(tmp_path: Path) -> Path:
    zip_base = Path(tmp_path, "zip_base")
    zip_base.mkdir()
    return zip_base


@pytest.fixture
def file_in_zip(zip_base_path: Path) -> Path:
    zipped_file = Path(zip_base_path, "zipped_file.txt")
    zipped_file.touch()
    return zipped_file


@pytest.fixture
def zipped_folder(zip_base_path: Path, file_in_zip: Path, tmp_path) -> Path:
    assert file_in_zip.exists()
    zip_out = Path(tmp_path, "zipped_folder")
    shutil.make_archive(zip_out.as_posix(), format="zip", root_dir=zip_base_path.as_posix())
    return Path(zip_out.as_posix() + ".zip")


@pytest.fixture
def destination_folder(tmp_path: Path) -> Path:
    return Path(tmp_path, "destination_folder")
